# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import mock

from tests import base

from st2client import shell
from st2client.models.core import ResourceManager
from st2client.models.action import Execution
from st2client.utils import httpclient

MOCK_MATCH_AND_EXECUTE_RESULT_1 = {
    "results": [
        {
            "execution": {
                "id": "mock-id",
            },
            "actionalias": {"ref": "mock-ref"},
        }
    ]
}

MOCK_MATCH_AND_EXECUTE_RESULT_2 = {
    "results": [
        {
            "execution": {"id": "mock-id-execute", "status": "succeeded"},
            "actionalias": {"ref": "mock-ref"},
            "liveaction": {
                "id": "mock-id",
            },
        }
    ]
}

MOCK_CREATE_EXECUTION_RESULT = {
    "id": "mock-id-format-execution",
    "status": "succeeded",
    "result": {"result": {"message": "Result formatted message"}},
}


class ActionAliasCommandTestCase(base.BaseCLITestCase):
    def __init__(self, *args, **kwargs):
        super(ActionAliasCommandTestCase, self).__init__(*args, **kwargs)
        self.shell = shell.Shell()

    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            return_value=base.FakeResponse(
                json.dumps(MOCK_MATCH_AND_EXECUTE_RESULT_1), 200, "OK"
            )
        ),
    )
    def test_match_and_execute_success(self):
        ret = self.shell.run(["action-alias", "execute", "run whoami on localhost"])
        self.assertEqual(ret, 0)

        expected_args = {
            "command": "run whoami on localhost",
            "user": "",
            "source_channel": "cli",
        }
        httpclient.HTTPClient.post.assert_called_with(
            "/aliasexecution/match_and_execute", expected_args
        )

        mock_stdout = self.stdout.getvalue()

        self.assertTrue("Matching Action-alias: 'mock-ref'" in mock_stdout)
        self.assertTrue("st2 execution get mock-id" in mock_stdout)

    @mock.patch.object(
        httpclient.HTTPClient,
        "post",
        mock.MagicMock(
            side_effect=[
                base.FakeResponse(
                    json.dumps(MOCK_MATCH_AND_EXECUTE_RESULT_2), 200, "OK"
                ),
                base.FakeResponse(json.dumps(MOCK_CREATE_EXECUTION_RESULT), 200, "OK"),
            ]
        ),
    )
    @mock.patch.object(
        ResourceManager,
        "get_by_id",
        mock.MagicMock(return_value=Execution(**MOCK_CREATE_EXECUTION_RESULT)),
    )
    def test_test_command_success(self):
        ret = self.shell.run(["action-alias", "test", "run whoami on localhost"])
        self.assertEqual(ret, 0)

        expected_args = {
            "command": "run whoami on localhost",
            "user": "",
            "source_channel": "cli",
        }
        httpclient.HTTPClient.post.assert_any_call(
            "/aliasexecution/match_and_execute", expected_args
        )

        expected_args = {
            "action": "chatops.format_execution_result",
            "parameters": {"execution_id": "mock-id-execute"},
            "user": "",
        }
        httpclient.HTTPClient.post.assert_any_call("/executions", expected_args)

        mock_stdout = self.stdout.getvalue()

        self.assertTrue(
            "Execution (mock-id-execute) has been started, waiting for it to finish"
            in mock_stdout
        )
        self.assertTrue(
            "Execution (mock-id-format-execution) has been started, waiting for it to "
            "finish" in mock_stdout
        )
        self.assertTrue("Formatted ChatOps result message" in mock_stdout)
        self.assertTrue("Result formatted message" in mock_stdout)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from st2client.commands import resource
from st2client.formatters import table
from st2client.models.webhook import Webhook


class WebhookBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(WebhookBranch, self).__init__(
            Webhook,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            read_only=True,
            commands={"list": WebhookListCommand, "get": WebhookGetCommand},
        )


class WebhookListCommand(resource.ContentPackResourceListCommand):
    display_attributes = ["url", "type", "description"]

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)

        for instance in instances:
            instance.url = instance.parameters["url"]

        instances = sorted(instances, key=lambda k: k.url)

        if args.json or args.yaml:
            self.print_output(
                instances,
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
                json=args.json,
                yaml=args.yaml,
            )
        else:
            self.print_output(
                instances,
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
            )


class WebhookGetCommand(resource.ResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = ["type", "description"]

    pk_argument_name = "url"



# -*- coding: utf-8 -*-
# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import time
import datetime
import json
import logging
import shutil
import tempfile

import requests
import six
import mock
import unittest2

import st2client
from st2client.shell import Shell
from st2client.client import Client
from st2client.utils import httpclient
from st2client.utils.misc import reencode_list_with_surrogate_escape_sequences
from st2common.models.db.auth import TokenDB
from tests import base

LOG = logging.getLogger(__name__)

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
CONFIG_FILE_PATH_FULL = os.path.join(BASE_DIR, "../fixtures/st2rc.full.ini")
CONFIG_FILE_PATH_PARTIAL = os.path.join(BASE_DIR, "../fixtures/st2rc.partial.ini")

MOCK_CONFIG = """
[credentials]
username = foo
password = bar
"""

MOCK_PACKAGE_METADATA = """
[server]
version = 2.8dev
git_sha = abcdefg
circle_build_url = https://circleci.com/gh/StackStorm/st2/7213
"""


class ShellTestCase(base.BaseCLITestCase):
    capture_output = True

    def __init__(self, *args, **kwargs):
        super(ShellTestCase, self).__init__(*args, **kwargs)
        self.shell = Shell()

    def setUp(self):
        super(ShellTestCase, self).setUp()

        if six.PY3:
            # In python --version outputs to stdout and in 2.x to stderr
            self.version_output = self.stdout
        else:
            self.version_output = self.stderr

    def test_commands_usage_and_help_strings(self):
        # No command, should print out user friendly usage / help string
        self.assertEqual(self.shell.run([]), 2)

        self.stderr.seek(0)
        stderr = self.stderr.read()
        self.assertIn("Usage: ", stderr)
        self.assertIn("For example:", stderr)
        self.assertIn("CLI for StackStorm", stderr)
        self.assertIn("positional arguments:", stderr)

        self.stdout.truncate()
        self.stderr.truncate()

        # --help should result in the same output
        try:
            self.assertEqual(self.shell.run(["--help"]), 0)
        except SystemExit as e:
            self.assertEqual(e.code, 0)

        self.stdout.seek(0)
        stdout = self.stdout.read()
        self.assertIn("Usage: ", stdout)
        self.assertIn("For example:", stdout)
        self.assertIn("CLI for StackStorm", stdout)
        self.assertIn("positional arguments:", stdout)

        self.stdout.truncate()
        self.stderr.truncate()

        # Sub command with no args
        try:
            self.assertEqual(self.shell.run(["action"]), 2)
        except SystemExit as e:
            self.assertEqual(e.code, 2)

        self.stderr.seek(0)
        stderr = self.stderr.read()

        self.assertIn("usage", stderr)

        if six.PY2:
            self.assertIn("{list,get,create,update", stderr)
            self.assertIn("error: too few arguments", stderr)

    def test_endpoints_default(self):
        base_url = "http://127.0.0.1"
        auth_url = "http://127.0.0.1:9100"
        api_url = "http://127.0.0.1:9101/v1"
        stream_url = "http://127.0.0.1:9102/v1"
        args = ["trigger", "list"]
        parsed_args = self.shell.parser.parse_args(args)
        client = self.shell.get_client(parsed_args)
        self.assertEqual(client.endpoints["base"], base_url)
        self.assertEqual(client.endpoints["auth"], auth_url)
        self.assertEqual(client.endpoints["api"], api_url)
        self.assertEqual(client.endpoints["stream"], stream_url)

    def test_endpoints_base_url_from_cli(self):
        base_url = "http://www.st2.com"
        auth_url = "http://www.st2.com:9100"
        api_url = "http://www.st2.com:9101/v1"
        stream_url = "http://www.st2.com:9102/v1"
        args = ["--url", base_url, "trigger", "list"]
        parsed_args = self.shell.parser.parse_args(args)
        client = self.shell.get_client(parsed_args)
        self.assertEqual(client.endpoints["base"], base_url)
        self.assertEqual(client.endpoints["auth"], auth_url)
        self.assertEqual(client.endpoints["api"], api_url)
        self.assertEqual(client.endpoints["stream"], stream_url)

    def test_endpoints_base_url_from_env(self):
        base_url = "http://www.st2.com"
        auth_url = "http://www.st2.com:9100"
        api_url = "http://www.st2.com:9101/v1"
        stream_url = "http://www.st2.com:9102/v1"
        os.environ["ST2_BASE_URL"] = base_url
        args = ["trigger", "list"]
        parsed_args = self.shell.parser.parse_args(args)
        client = self.shell.get_client(parsed_args)
        self.assertEqual(client.endpoints["base"], base_url)
        self.assertEqual(client.endpoints["auth"], auth_url)
        self.assertEqual(client.endpoints["api"], api_url)
        self.assertEqual(client.endpoints["stream"], stream_url)

    def test_endpoints_override_from_cli(self):
        base_url = "http://www.st2.com"
        auth_url = "http://www.st2.com:8888"
        api_url = "http://www.stackstorm1.com:9101/v1"
        stream_url = "http://www.stackstorm1.com:9102/v1"
        args = [
            "--url",
            base_url,
            "--auth-url",
            auth_url,
            "--api-url",
            api_url,
            "--stream-url",
            stream_url,
            "trigger",
            "list",
        ]
        parsed_args = self.shell.parser.parse_args(args)
        client = self.shell.get_client(parsed_args)
        self.assertEqual(client.endpoints["base"], base_url)
        self.assertEqual(client.endpoints["auth"], auth_url)
        self.assertEqual(client.endpoints["api"], api_url)
        self.assertEqual(client.endpoints["stream"], stream_url)

    def test_endpoints_override_from_env(self):
        base_url = "http://www.st2.com"
        auth_url = "http://www.st2.com:8888"
        api_url = "http://www.stackstorm1.com:9101/v1"
        stream_url = "http://www.stackstorm1.com:9102/v1"
        os.environ["ST2_BASE_URL"] = base_url
        os.environ["ST2_AUTH_URL"] = auth_url
        os.environ["ST2_API_URL"] = api_url
        os.environ["ST2_STREAM_URL"] = stream_url
        args = ["trigger", "list"]
        parsed_args = self.shell.parser.parse_args(args)
        client = self.shell.get_client(parsed_args)
        self.assertEqual(client.endpoints["base"], base_url)
        self.assertEqual(client.endpoints["auth"], auth_url)
        self.assertEqual(client.endpoints["api"], api_url)
        self.assertEqual(client.endpoints["stream"], stream_url)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES), 200, "OK")
        ),
    )
    def test_exit_code_on_success(self):
        argv = ["trigger", "list"]
        self.assertEqual(self.shell.run(argv), 0)

    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(None, 500, "INTERNAL SERVER ERROR")
        ),
    )
    def test_exit_code_on_error(self):
        argv = ["trigger", "list"]
        self.assertEqual(self.shell.run(argv), 1)

    def _validate_parser(self, args_list, is_subcommand=True):
        for args in args_list:
            ns = self.shell.parser.parse_args(args)
            func = (
                self.shell.commands[args[0]].run_and_print
                if not is_subcommand
                else self.shell.commands[args[0]].commands[args[1]].run_and_print
            )
            self.assertEqual(ns.func, func)

    def test_action(self):
        args_list = [
            ["action", "list"],
            ["action", "get", "abc"],
            ["action", "create", "/tmp/action.json"],
            ["action", "update", "123", "/tmp/action.json"],
            ["action", "delete", "abc"],
            ["action", "execute", "-h"],
            ["action", "execute", "remote", "-h"],
            [
                "action",
                "execute",
                "remote",
                "hosts=192.168.1.1",
                "user=st2",
                'cmd="ls -l"',
            ],
            ["action", "execute", "remote-fib", "hosts=192.168.1.1", "3", "8"],
        ]
        self._validate_parser(args_list)

    def test_action_execution(self):
        args_list = [
            ["execution", "list"],
            ["execution", "list", "-a", "all"],
            ["execution", "list", "--attr=all"],
            ["execution", "get", "123"],
            ["execution", "get", "123", "-d"],
            ["execution", "get", "123", "-k", "localhost.stdout"],
            ["execution", "re-run", "123"],
            ["execution", "re-run", "123", "--tasks", "x", "y", "z"],
            ["execution", "re-run", "123", "--tasks", "x", "y", "z", "--no-reset", "x"],
            ["execution", "re-run", "123", "a=1", "b=x", "c=True"],
            ["execution", "cancel", "123"],
            ["execution", "cancel", "123", "456"],
            ["execution", "pause", "123"],
            ["execution", "pause", "123", "456"],
            ["execution", "resume", "123"],
            ["execution", "resume", "123", "456"],
        ]
        self._validate_parser(args_list)

        # Test mutually exclusive argument groups
        self.assertRaises(
            SystemExit,
            self._validate_parser,
            [["execution", "get", "123", "-d", "-k", "localhost.stdout"]],
        )

    def test_key(self):
        args_list = [
            ["key", "list"],
            ["key", "list", "-n", "2"],
            ["key", "get", "abc"],
            ["key", "set", "abc", "123"],
            ["key", "delete", "abc"],
            ["key", "load", "/tmp/keys.json"],
        ]
        self._validate_parser(args_list)

    def test_policy(self):
        args_list = [
            ["policy", "list"],
            ["policy", "list", "-p", "core"],
            ["policy", "list", "--pack", "core"],
            ["policy", "list", "-r", "core.local"],
            ["policy", "list", "--resource-ref", "core.local"],
            ["policy", "list", "-pt", "action.type1"],
            ["policy", "list", "--policy-type", "action.type1"],
            ["policy", "list", "-r", "core.local", "-pt", "action.type1"],
            [
                "policy",
                "list",
                "--resource-ref",
                "core.local",
                "--policy-type",
                "action.type1",
            ],
            ["policy", "get", "abc"],
            ["policy", "create", "/tmp/policy.json"],
            ["policy", "update", "123", "/tmp/policy.json"],
            ["policy", "delete", "abc"],
        ]
        self._validate_parser(args_list)

    def test_policy_type(self):
        args_list = [
            ["policy-type", "list"],
            ["policy-type", "list", "-r", "action"],
            ["policy-type", "list", "--resource-type", "action"],
            ["policy-type", "get", "abc"],
        ]
        self._validate_parser(args_list)

    def test_pack(self):
        args_list = [
            ["pack", "list"],
            ["pack", "get", "abc"],
            ["pack", "search", "abc"],
            ["pack", "show", "abc"],
            ["pack", "remove", "abc"],
            ["pack", "remove", "abc", "--detail"],
            ["pack", "install", "abc"],
            ["pack", "install", "abc", "--force"],
            ["pack", "install", "abc", "--detail"],
            ["pack", "config", "abc"],
        ]
        self._validate_parser(args_list)

    @mock.patch("st2client.base.ST2_CONFIG_PATH", "/home/does/not/exist")
    def test_print_config_default_config_no_config(self):
        os.environ["ST2_CONFIG_FILE"] = "/home/does/not/exist"
        argv = ["--print-config"]
        self.assertEqual(self.shell.run(argv), 3)

        self.stdout.seek(0)
        stdout = self.stdout.read()

        self.assertIn("username = None", stdout)
        self.assertIn("cache_token = True", stdout)

    def test_print_config_custom_config_as_env_variable(self):
        os.environ["ST2_CONFIG_FILE"] = CONFIG_FILE_PATH_FULL
        argv = ["--print-config"]
        self.assertEqual(self.shell.run(argv), 3)

        self.stdout.seek(0)
        stdout = self.stdout.read()

        self.assertIn("username = test1", stdout)
        self.assertIn("cache_token = False", stdout)

    def test_print_config_custom_config_as_command_line_argument(self):
        argv = ["--print-config", "--config-file=%s" % (CONFIG_FILE_PATH_FULL)]
        self.assertEqual(self.shell.run(argv), 3)

        self.stdout.seek(0)
        stdout = self.stdout.read()

        self.assertIn("username = test1", stdout)
        self.assertIn("cache_token = False", stdout)

    def test_run(self):
        args_list = [
            ["run", "-h"],
            ["run", "abc", "-h"],
            ["run", "remote", "hosts=192.168.1.1", "user=st2", 'cmd="ls -l"'],
            ["run", "remote-fib", "hosts=192.168.1.1", "3", "8"],
        ]
        self._validate_parser(args_list, is_subcommand=False)

    def test_runner(self):
        args_list = [["runner", "list"], ["runner", "get", "abc"]]
        self._validate_parser(args_list)

    def test_rule(self):
        args_list = [
            ["rule", "list"],
            ["rule", "list", "-n", "1"],
            ["rule", "get", "abc"],
            ["rule", "create", "/tmp/rule.json"],
            ["rule", "update", "123", "/tmp/rule.json"],
            ["rule", "delete", "abc"],
        ]
        self._validate_parser(args_list)

    def test_trigger(self):
        args_list = [
            ["trigger", "list"],
            ["trigger", "get", "abc"],
            ["trigger", "create", "/tmp/trigger.json"],
            ["trigger", "update", "123", "/tmp/trigger.json"],
            ["trigger", "delete", "abc"],
        ]
        self._validate_parser(args_list)

    def test_workflow(self):
        args_list = [
            ["workflow", "inspect", "--file", "/path/to/workflow/definition"],
            ["workflow", "inspect", "--action", "mock.foobar"],
        ]

        self._validate_parser(args_list)

    @mock.patch("sys.exit", mock.Mock())
    @mock.patch("st2client.shell.__version__", "v2.8.0")
    def test_get_version_no_package_metadata_file_stable_version(self):
        # stable version, package metadata file doesn't exist on disk - no git revision should be
        # included
        shell = Shell()
        shell.parser.parse_args(args=["--version"])

        self.version_output.seek(0)
        stderr = self.version_output.read()
        self.assertIn("v2.8.0, on Python", stderr)

    @mock.patch("sys.exit", mock.Mock())
    @mock.patch("st2client.shell.__version__", "v2.8.0")
    def test_get_version_package_metadata_file_exists_stable_version(self):
        # stable version, package metadata file exists on disk - no git revision should be included
        package_metadata_path = self._write_mock_package_metadata_file()
        st2client.shell.PACKAGE_METADATA_FILE_PATH = package_metadata_path

        shell = Shell()
        shell.run(argv=["--version"])

        self.version_output.seek(0)
        stderr = self.version_output.read()
        self.assertIn("v2.8.0, on Python", stderr)

    @mock.patch("sys.exit", mock.Mock())
    @mock.patch("st2client.shell.__version__", "v2.9dev")
    @mock.patch("st2client.shell.PACKAGE_METADATA_FILE_PATH", "/tmp/doesnt/exist.1")
    def test_get_version_no_package_metadata_file_dev_version(self):
        # dev version, package metadata file doesn't exist on disk - no git revision should be
        # included since package metadata file doesn't exist on disk
        shell = Shell()
        shell.parser.parse_args(args=["--version"])

        self.version_output.seek(0)
        stderr = self.version_output.read()
        self.assertIn("v2.9dev, on Python", stderr)

    @mock.patch("sys.exit", mock.Mock())
    @mock.patch("st2client.shell.__version__", "v2.9dev")
    def test_get_version_package_metadata_file_exists_dev_version(self):
        # dev version, package metadata file exists on disk - git revision should be included
        # since package metadata file exists on disk and contains server.git_sha attribute
        package_metadata_path = self._write_mock_package_metadata_file()
        st2client.shell.PACKAGE_METADATA_FILE_PATH = package_metadata_path

        shell = Shell()
        shell.parser.parse_args(args=["--version"])

        self.version_output.seek(0)
        stderr = self.version_output.read()
        self.assertIn("v2.9dev (abcdefg), on Python", stderr)

    @mock.patch("locale.getdefaultlocale", mock.Mock(return_value=["en_US"]))
    @mock.patch("locale.getpreferredencoding", mock.Mock(return_value="iso"))
    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES), 200, "OK")
        ),
    )
    @mock.patch("st2client.shell.LOGGER")
    def test_non_unicode_encoding_locale_warning_is_printed(self, mock_logger):
        shell = Shell()
        shell.run(argv=["trigger", "list"])

        call_args = mock_logger.warn.call_args[0][0]
        self.assertIn(
            "Locale en_US with encoding iso which is not UTF-8 is used.", call_args
        )

    @mock.patch("locale.getdefaultlocale", mock.Mock(side_effect=ValueError("bar")))
    @mock.patch("locale.getpreferredencoding", mock.Mock(side_effect=ValueError("bar")))
    @mock.patch.object(
        httpclient.HTTPClient,
        "get",
        mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES), 200, "OK")
        ),
    )
    @mock.patch("st2client.shell.LOGGER")
    def test_failed_to_get_locale_encoding_warning_is_printed(self, mock_logger):
        shell = Shell()
        shell.run(argv=["trigger", "list"])

        call_args = mock_logger.warn.call_args[0][0]
        self.assertTrue(
            "Locale unknown with encoding unknown which is not UTF-8 is used."
            in call_args
        )

    def _write_mock_package_metadata_file(self):
        _, package_metadata_path = tempfile.mkstemp()

        with open(package_metadata_path, "w") as fp:
            fp.write(MOCK_PACKAGE_METADATA)

        return package_metadata_path

    @unittest2.skipIf(True, "skipping until checks are re-enabled")
    @mock.patch.object(
        requests, "get", mock.MagicMock(return_value=base.FakeResponse("{}", 200, "OK"))
    )
    def test_dont_warn_multiple_times(self):
        mock_temp_dir_path = tempfile.mkdtemp()
        mock_config_dir_path = os.path.join(mock_temp_dir_path, "testconfig")
        mock_config_path = os.path.join(mock_config_dir_path, "config")

        # Make the temporary config directory
        os.makedirs(mock_config_dir_path)

        old_perms = os.stat(mock_config_dir_path).st_mode
        new_perms = old_perms | 0o7
        os.chmod(mock_config_dir_path, new_perms)

        # Make the temporary config file
        shutil.copyfile(CONFIG_FILE_PATH_FULL, mock_config_path)
        os.chmod(mock_config_path, 0o777)  # nosec

        shell = Shell()
        shell.LOG = mock.Mock()

        # Test without token.
        shell.run(["--config-file", mock_config_path, "action", "list"])

        self.assertEqual(shell.LOG.warn.call_count, 2)
        self.assertEqual(
            shell.LOG.warn.call_args_list[0][0][0][:63],
            "The StackStorm configuration directory permissions are insecure",
        )
        self.assertEqual(
            shell.LOG.warn.call_args_list[1][0][0][:58],
            "The StackStorm configuration file permissions are insecure",
        )

        self.assertEqual(shell.LOG.info.call_count, 2)
        self.assertEqual(
            shell.LOG.info.call_args_list[0][0][0],
            "The SGID bit is not " "set on the StackStorm configuration directory.",
        )

        self.assertEqual(
            shell.LOG.info.call_args_list[1][0][0], "Skipping parsing CLI config"
        )

    def test_policy_list_with_pack_option(self):
        argv = ["policy", "list", "-p", "test"]
        mock_obj = mock.MagicMock(
            return_value=base.FakeResponse(json.dumps(base.RESOURCES), 200, "OK")
        )
        with mock.patch.object(httpclient.HTTPClient, "get", mock_obj):
            self.shell.run(argv)
            self.assertEqual(
                mock_obj.mock_calls[0],
                mock.call(
                    "/policies/?include_attributes=ref%2Cresource_ref%2C"
                    "policy_type%2Cenabled&pack=test"
                ),
            )


class CLITokenCachingTestCase(unittest2.TestCase):
    def setUp(self):
        super(CLITokenCachingTestCase, self).setUp()
        self._mock_temp_dir_path = tempfile.mkdtemp()
        self._mock_config_directory_path = os.path.join(
            self._mock_temp_dir_path, "testconfig"
        )
        self._mock_config_path = os.path.join(
            self._mock_config_directory_path, "config"
        )

        os.makedirs(self._mock_config_directory_path)

        self._p1 = mock.patch(
            "st2client.base.ST2_CONFIG_DIRECTORY", self._mock_config_directory_path
        )
        self._p2 = mock.patch("st2client.base.ST2_CONFIG_PATH", self._mock_config_path)
        self._p1.start()
        self._p2.start()

    def tearDown(self):
        super(CLITokenCachingTestCase, self).tearDown()
        self._p1.stop()
        self._p2.stop()

        for var in [
            "ST2_BASE_URL",
            "ST2_API_URL",
            "ST2_STREAM_URL",
            "ST2_DATASTORE_URL",
            "ST2_AUTH_TOKEN",
        ]:
            if var in os.environ:
                del os.environ[var]

    def _write_mock_config(self):
        with open(self._mock_config_path, "w") as fp:
            fp.write(MOCK_CONFIG)

    def test_get_cached_auth_token_invalid_permissions(self):
        shell = Shell()
        client = Client()
        username = "testu"
        password = "testp"

        cached_token_path = shell._get_cached_token_path_for_user(username=username)
        data = {"token": "yayvalid", "expire_timestamp": (int(time.time()) + 20)}
        with open(cached_token_path, "w") as fp:
            fp.write(json.dumps(data))

        # 1. Current user doesn't have read access to the config directory
        os.chmod(self._mock_config_directory_path, 0o000)

        shell.LOG = mock.Mock()
        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )

        self.assertEqual(result, None)
        self.assertEqual(shell.LOG.warn.call_count, 1)
        log_message = shell.LOG.warn.call_args[0][0]

        expected_msg = (
            "Unable to retrieve cached token from .*? read access to the parent "
            "directory"
        )
        self.assertRegexpMatches(log_message, expected_msg)

        # 2. Read access on the directory, but not on the cached token file
        os.chmod(self._mock_config_directory_path, 0o777)  # nosec
        os.chmod(cached_token_path, 0o000)

        shell.LOG = mock.Mock()
        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )
        self.assertEqual(result, None)

        self.assertEqual(shell.LOG.warn.call_count, 1)
        log_message = shell.LOG.warn.call_args[0][0]

        expected_msg = (
            "Unable to retrieve cached token from .*? read access to this file"
        )
        self.assertRegexpMatches(log_message, expected_msg)

        # 3. Other users also have read access to the file
        os.chmod(self._mock_config_directory_path, 0o777)  # nosec
        os.chmod(cached_token_path, 0o444)

        shell.LOG = mock.Mock()
        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )
        self.assertEqual(result, "yayvalid")

        self.assertEqual(shell.LOG.warn.call_count, 1)
        log_message = shell.LOG.warn.call_args[0][0]

        expected_msg = "Permissions .*? for cached token file .*? are too permissive.*"
        self.assertRegexpMatches(log_message, expected_msg)

    def test_cache_auth_token_invalid_permissions(self):
        shell = Shell()
        username = "testu"

        cached_token_path = shell._get_cached_token_path_for_user(username=username)
        expiry = datetime.datetime.utcnow() + datetime.timedelta(seconds=30)

        token_db = TokenDB(user=username, token="fyeah", expiry=expiry)

        cached_token_path = shell._get_cached_token_path_for_user(username=username)
        data = {"token": "yayvalid", "expire_timestamp": (int(time.time()) + 20)}
        with open(cached_token_path, "w") as fp:
            fp.write(json.dumps(data))

        # 1. Current user has no write access to the parent directory
        os.chmod(self._mock_config_directory_path, 0o000)

        shell.LOG = mock.Mock()
        shell._cache_auth_token(token_obj=token_db)

        self.assertEqual(shell.LOG.warn.call_count, 1)
        log_message = shell.LOG.warn.call_args[0][0]

        expected_msg = (
            "Unable to write token to .*? doesn't have write access to the parent "
            "directory"
        )
        self.assertRegexpMatches(log_message, expected_msg)

        # 2. Current user has no write access to the cached token file
        os.chmod(self._mock_config_directory_path, 0o777)  # nosec
        os.chmod(cached_token_path, 0o000)

        shell.LOG = mock.Mock()
        shell._cache_auth_token(token_obj=token_db)

        self.assertEqual(shell.LOG.warn.call_count, 1)
        log_message = shell.LOG.warn.call_args[0][0]

        expected_msg = (
            "Unable to write token to .*? doesn't have write access to this file"
        )
        self.assertRegexpMatches(log_message, expected_msg)

    def test_get_cached_auth_token_no_token_cache_file(self):
        client = Client()
        shell = Shell()
        username = "testu"
        password = "testp"

        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )
        self.assertEqual(result, None)

    def test_get_cached_auth_token_corrupted_token_cache_file(self):
        client = Client()
        shell = Shell()
        username = "testu"
        password = "testp"

        cached_token_path = shell._get_cached_token_path_for_user(username=username)
        with open(cached_token_path, "w") as fp:
            fp.write("CORRRRRUPTED!")

        expected_msg = "File (.+) with cached token is corrupted or invalid"
        self.assertRaisesRegexp(
            ValueError,
            expected_msg,
            shell._get_cached_auth_token,
            client=client,
            username=username,
            password=password,
        )

    def test_get_cached_auth_token_expired_token_in_cache_file(self):
        client = Client()
        shell = Shell()
        username = "testu"
        password = "testp"

        cached_token_path = shell._get_cached_token_path_for_user(username=username)
        data = {"token": "expired", "expire_timestamp": (int(time.time()) - 10)}
        with open(cached_token_path, "w") as fp:
            fp.write(json.dumps(data))

        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )
        self.assertEqual(result, None)

    def test_get_cached_auth_token_valid_token_in_cache_file(self):
        client = Client()
        shell = Shell()
        username = "testu"
        password = "testp"

        cached_token_path = shell._get_cached_token_path_for_user(username=username)
        data = {"token": "yayvalid", "expire_timestamp": (int(time.time()) + 20)}
        with open(cached_token_path, "w") as fp:
            fp.write(json.dumps(data))

        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )
        self.assertEqual(result, "yayvalid")

    def test_cache_auth_token_success(self):
        client = Client()
        shell = Shell()
        username = "testu"
        password = "testp"
        expiry = datetime.datetime.utcnow() + datetime.timedelta(seconds=30)

        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )
        self.assertEqual(result, None)

        token_db = TokenDB(user=username, token="fyeah", expiry=expiry)
        shell._cache_auth_token(token_obj=token_db)

        result = shell._get_cached_auth_token(
            client=client, username=username, password=password
        )
        self.assertEqual(result, "fyeah")

    def test_automatic_auth_skipped_on_auth_command(self):
        self._write_mock_config()

        shell = Shell()
        shell._get_auth_token = mock.Mock()

        argv = ["auth", "testu", "-p", "testp"]
        args = shell.parser.parse_args(args=argv)
        shell.get_client(args=args)
        self.assertEqual(shell._get_auth_token.call_count, 0)

    def test_automatic_auth_skipped_if_token_provided_as_env_variable(self):
        self._write_mock_config()

        shell = Shell()
        shell._get_auth_token = mock.Mock()

        os.environ["ST2_AUTH_TOKEN"] = "fooo"
        argv = ["action", "list"]
        args = shell.parser.parse_args(args=argv)
        shell.get_client(args=args)
        self.assertEqual(shell._get_auth_token.call_count, 0)

    def test_automatic_auth_skipped_if_token_provided_as_cli_argument(self):
        self._write_mock_config()

        shell = Shell()
        shell._get_auth_token = mock.Mock()

        argv = ["action", "list", "--token=bar"]
        args = shell.parser.parse_args(args=argv)
        shell.get_client(args=args)
        self.assertEqual(shell._get_auth_token.call_count, 0)

        argv = ["action", "list", "-t", "bar"]
        args = shell.parser.parse_args(args=argv)
        shell.get_client(args=args)
        self.assertEqual(shell._get_auth_token.call_count, 0)

    def test_get_one_unicode_character_in_name(self):
        self._write_mock_config()

        shell = Shell()
        shell._get_auth_token = mock.Mock()

        os.environ["ST2_AUTH_TOKEN"] = "fooo"
        argv = ["action", "get", "examples.test_rule_utf8_náme"]
        args = shell.parser.parse_args(args=argv)
        shell.get_client(args=args)
        self.assertEqual(args.ref_or_id, ["examples.test_rule_utf8_náme"])

    def test_reencode_list_replace_surrogate_escape(self):
        value = ["a", "b", "c", "d"]
        expected = value
        result = reencode_list_with_surrogate_escape_sequences(value)

        self.assertEqual(result, expected)

        value = ["a", "b", "c", "n\udcc3\udca1me"]
        expected = ["a", "b", "c", "náme"]
        result = reencode_list_with_surrogate_escape_sequences(value)

        self.assertEqual(result, expected)

        value = ["a", "b", "c", "你好"]
        expected = value
        result = reencode_list_with_surrogate_escape_sequences(value)

        self.assertEqual(result, expected)



# -*- coding: utf-8 -*-
# NOTE: This file is auto-generated - DO NOT EDIT MANUALLY
#       Instead modify scripts/dist_utils.py and run 'make .sdist-requirements' to
#       update dist_utils.py files for all components

# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import re
import sys

# NOTE: This script can't rely on any 3rd party dependency so we need to use this code here
#
# TODO: Why can't this script rely on 3rd party dependencies? Is it because it has to import
#       from pip?
#
# TODO: Dear future developer, if you are back here fixing a bug with how we parse
#       requirements files, please look into using the packaging package on PyPI:
#       https://packaging.pypa.io/en/latest/requirements/
#       and specifying that in the `setup_requires` argument to `setuptools.setup()`
#       for subpackages.
#       At the very least we can vendorize some of their code instead of reimplementing
#       each piece of their code every time our parsing breaks.
PY3 = sys.version_info[0] == 3

if PY3:
    text_type = str
else:
    text_type = unicode  # noqa  # pylint: disable=E0602

GET_PIP = "curl https://bootstrap.pypa.io/get-pip.py | python"

__all__ = [
    "fetch_requirements",
    "apply_vagrant_workaround",
    "get_version_string",
    "parse_version_string",
]


def fetch_requirements(requirements_file_path):
    """
    Return a list of requirements and links by parsing the provided requirements file.
    """
    links = []
    reqs = []

    def _get_link(line):
        vcs_prefixes = ["git+", "svn+", "hg+", "bzr+"]

        for vcs_prefix in vcs_prefixes:
            if line.startswith(vcs_prefix) or line.startswith("-e %s" % (vcs_prefix)):
                req_name = re.findall(".*#egg=(.+)([&|@]).*$", line)

                if not req_name:
                    req_name = re.findall(".*#egg=(.+?)$", line)
                else:
                    req_name = req_name[0]

                if not req_name:
                    raise ValueError(
                        'Line "%s" is missing "#egg=<package name>"' % (line)
                    )

                link = line.replace("-e ", "").strip()
                return link, req_name[0]
            elif vcs_prefix in line and line.count("@") == 2:
                # PEP 440 direct reference: <package name>@ <url>@version
                req_name, link = line.split("@", 1)
                req_name = req_name.strip()
                link = f"{link.strip()}#egg={req_name}"
                return link, req_name

        return None, None

    with open(requirements_file_path, "r") as fp:
        for line in fp.readlines():
            line = line.strip()

            if line.startswith("#") or not line:
                continue

            link, req_name = _get_link(line=line)

            if link:
                links.append(link)
            else:
                req_name = line

                if ";" in req_name:
                    req_name = req_name.split(";")[0].strip()

            reqs.append(req_name)

    return (reqs, links)


def apply_vagrant_workaround():
    """
    Function which detects if the script is being executed inside vagrant and if it is, it deletes
    "os.link" attribute.
    Note: Without this workaround, setup.py sdist will fail when running inside a shared directory
    (nfs / virtualbox shared folders).
    """
    if os.environ.get("USER", None) == "vagrant":
        del os.link


def get_version_string(init_file):
    """
    Read __version__ string for an init file.
    """

    with open(init_file, "r") as fp:
        content = fp.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M)
        if version_match:
            return version_match.group(1)

        raise RuntimeError("Unable to find version string in %s." % (init_file))


# alias for get_version_string
parse_version_string = get_version_string



#!/usr/bin/env python

# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Command-line interface to StackStorm.
"""

from __future__ import print_function
from __future__ import absolute_import

# Ignore CryptographyDeprecationWarning warnings which appear on Python 3.6
# TODO: Remove after dropping python3.6
import warnings

warnings.filterwarnings("ignore", message="Python 3.6 is no longer supported")

import os
import sys
import argcomplete
import argparse
import logging
import locale

import six
from six.moves.configparser import ConfigParser

from st2client import __version__
from st2client import models
from st2client.base import BaseCLIApp
from st2client.commands import auth
from st2client.commands import action
from st2client.commands import action_alias
from st2client.commands import keyvalue
from st2client.commands import inquiry
from st2client.commands import pack
from st2client.commands import policy
from st2client.commands import resource
from st2client.commands import sensor
from st2client.commands import trace
from st2client.commands import trigger
from st2client.commands import triggerinstance
from st2client.commands import timer
from st2client.commands import webhook
from st2client.commands import rule
from st2client.commands import rule_enforcement
from st2client.commands import rbac
from st2client.commands import workflow
from st2client.commands import service_registry
from st2client.config import set_config
from st2client.exceptions.operations import OperationFailureException
from st2client.utils.logging import LogLevelFilter, set_log_level_for_all_loggers
from st2client.utils.misc import reencode_list_with_surrogate_escape_sequences
from st2client.commands.auth import TokenCreateCommand
from st2client.commands.auth import LoginCommand
from st2client.utils.profiler import setup_regular_profiler

__all__ = ["Shell"]

LOGGER = logging.getLogger(__name__)

CLI_DESCRIPTION = (
    "CLI for StackStorm event-driven automation platform. https://stackstorm.com"
)
USAGE_STRING = """
Usage: %(prog)s [options] <command> <sub command> [options]

For example:

    %(prog)s action list --pack=st2
    %(prog)s run core.local cmd=date
    %(prog)s --debug run core.local cmd=date
""".strip()

NON_UTF8_LOCALE = (
    """
Locale %s with encoding %s which is not UTF-8 is used. This means that some functionality which
relies on outputting unicode characters won't work.

You are encouraged to use UTF-8 locale by setting LC_ALL environment variable to en_US.UTF-8 or
similar.
""".strip()
    .replace("\n", " ")
    .replace("  ", " ")
)

PACKAGE_METADATA_FILE_PATH = "/opt/stackstorm/st2/package.meta"

"""
Here we sanitize the provided args and ensure they contain valid unicode values.

By default, sys.argv will contain a unicode string where the actual item values which contain
unicode sequences are escaped using unicode surrogates.

For example, if "examples.test_rule_utf8_náme" value is specified as a CLI argument, sys.argv
and as such also url, would contain "examples.test_rule_utf8_n%ED%B3%83%ED%B2%A1me" which is not
what we want.

Complete sys.argv example:

1. Default - ['shell.py', '--debug', 'rule', 'get', 'examples.test_rule_utf8_n\udcc3\udca1me']
2. What we want - ['shell.py', '--debug', 'rule', 'get', 'examples.test_rule_utf8_náme']

This won't work correctly when sending requests to the API. As such, we correctly escape the
value to the unicode string here and then let the http layer (requests) correctly url encode
this value.

Technically, we could also just try to re-encode it in the HTTPClient and I tried that first, but
it turns out more code in the client results in exceptions if it's not re-encoded as early as
possible.
"""

REENCODE_ARGV = os.environ.get("ST2_CLI_RENCODE_ARGV", "true").lower() in [
    "true",
    "1",
    "yes",
]

if REENCODE_ARGV:
    try:
        sys.argv = reencode_list_with_surrogate_escape_sequences(sys.argv)
    except Exception as e:
        print("Failed to re-encode sys.argv: %s" % (str(e)))


def get_stackstorm_version():
    """
    Return StackStorm version including git commit revision if running a dev release and a file
    with package metadata which includes git revision is available.

    :rtype: ``str``
    """
    if "dev" in __version__:
        version = __version__

        if not os.path.isfile(PACKAGE_METADATA_FILE_PATH):
            return version

        config = ConfigParser()

        try:
            config.read(PACKAGE_METADATA_FILE_PATH)
        except Exception:
            return version

        try:
            git_revision = config.get("server", "git_sha")
        except Exception:
            return version

        version = "%s (%s)" % (version, git_revision)
    else:
        version = __version__

    return version


class Shell(BaseCLIApp):
    LOG = LOGGER

    SKIP_AUTH_CLASSES = [
        TokenCreateCommand.__name__,
        LoginCommand.__name__,
    ]

    def __init__(self):
        # Set up of endpoints is delayed until program is run.
        self.client = None

        # Set up the main parser.
        self.parser = argparse.ArgumentParser(description=CLI_DESCRIPTION)

        # Set up general program options.
        self.parser.add_argument(
            "--version",
            action="version",
            version="%(prog)s {version}, on Python {python_major}.{python_minor}.{python_patch}".format(
                version=get_stackstorm_version(),
                python_major=sys.version_info.major,
                python_minor=sys.version_info.minor,
                python_patch=sys.version_info.micro,
            ),
        )

        self.parser.add_argument(
            "--url",
            action="store",
            dest="base_url",
            default=None,
            help="Base URL for the API servers. Assumes all servers use the "
            "same base URL and default ports are used. Get ST2_BASE_URL "
            "from the environment variables by default.",
        )

        self.parser.add_argument(
            "--auth-url",
            action="store",
            dest="auth_url",
            default=None,
            help="URL for the authentication service. Get ST2_AUTH_URL "
            "from the environment variables by default.",
        )

        self.parser.add_argument(
            "--api-url",
            action="store",
            dest="api_url",
            default=None,
            help="URL for the API server. Get ST2_API_URL "
            "from the environment variables by default.",
        )

        self.parser.add_argument(
            "--stream-url",
            action="store",
            dest="stream_url",
            default=None,
            help="URL for the stream endpoint. Get ST2_STREAM_URL"
            "from the environment variables by default.",
        )

        self.parser.add_argument(
            "--api-version",
            action="store",
            dest="api_version",
            default=None,
            help="API version to use. Get ST2_API_VERSION "
            "from the environment variables by default.",
        )

        self.parser.add_argument(
            "--cacert",
            action="store",
            dest="cacert",
            default=None,
            help="Path to the CA cert bundle for the SSL endpoints. "
            "Get ST2_CACERT from the environment variables by default. "
            "If this is not provided, then SSL cert will not be verified.",
        )

        self.parser.add_argument(
            "--basic-auth",
            action="store",
            dest="basic_auth",
            default=None,
            help="Optional additional basic auth credentials used to authenticate",
        )

        self.parser.add_argument(
            "--config-file",
            action="store",
            dest="config_file",
            default=None,
            help="Path to the CLI config file",
        )

        self.parser.add_argument(
            "--print-config",
            action="store_true",
            dest="print_config",
            default=False,
            help="Parse the config file and print the values",
        )

        self.parser.add_argument(
            "--skip-config",
            action="store_true",
            dest="skip_config",
            default=False,
            help="Don't parse and use the CLI config file",
        )

        self.parser.add_argument(
            "--debug",
            action="store_true",
            dest="debug",
            default=False,
            help="Enable debug mode",
        )

        # Set up list of commands and subcommands.
        self.subparsers = self.parser.add_subparsers(dest="parser")
        self.subparsers.required = True
        self.commands = {}

        self.commands["run"] = action.ActionRunCommand(
            models.Action, self, self.subparsers, name="run", add_help=False
        )

        self.commands["action"] = action.ActionBranch(
            "An activity that happens as a response to the external event.",
            self,
            self.subparsers,
        )

        self.commands["action-alias"] = action_alias.ActionAliasBranch(
            "Action aliases.", self, self.subparsers
        )

        self.commands["auth"] = auth.TokenCreateCommand(
            models.Token, self, self.subparsers, name="auth"
        )

        self.commands["login"] = auth.LoginCommand(
            models.Token, self, self.subparsers, name="login"
        )

        self.commands["whoami"] = auth.WhoamiCommand(
            models.Token, self, self.subparsers, name="whoami"
        )

        self.commands["api-key"] = auth.ApiKeyBranch("API Keys.", self, self.subparsers)

        self.commands["execution"] = action.ActionExecutionBranch(
            "An invocation of an action.", self, self.subparsers
        )

        self.commands["inquiry"] = inquiry.InquiryBranch(
            "Inquiries provide an opportunity to ask a question "
            "and wait for a response in a workflow.",
            self,
            self.subparsers,
        )

        self.commands["key"] = keyvalue.KeyValuePairBranch(
            "Key value pair is used to store commonly used configuration "
            "for reuse in sensors, actions, and rules.",
            self,
            self.subparsers,
        )

        self.commands["pack"] = pack.PackBranch(
            "A group of related integration resources: " "actions, rules, and sensors.",
            self,
            self.subparsers,
        )

        self.commands["policy"] = policy.PolicyBranch(
            "Policy that is enforced on a resource.", self, self.subparsers
        )

        self.commands["policy-type"] = policy.PolicyTypeBranch(
            "Type of policy that can be applied to resources.", self, self.subparsers
        )

        self.commands["rule"] = rule.RuleBranch(
            'A specification to invoke an "action" on a "trigger" selectively '
            "based on some criteria.",
            self,
            self.subparsers,
        )

        self.commands["webhook"] = webhook.WebhookBranch(
            "Webhooks.", self, self.subparsers
        )

        self.commands["timer"] = timer.TimerBranch("Timers.", self, self.subparsers)

        self.commands["runner"] = resource.ResourceBranch(
            models.RunnerType,
            "Runner is a type of handler for a specific class of actions.",
            self,
            self.subparsers,
            read_only=True,
            has_disable=True,
        )

        self.commands["sensor"] = sensor.SensorBranch(
            "An adapter which allows you to integrate StackStorm with external system.",
            self,
            self.subparsers,
        )

        self.commands["trace"] = trace.TraceBranch(
            "A group of executions, rules and triggerinstances that are related.",
            self,
            self.subparsers,
        )

        self.commands["trigger"] = trigger.TriggerTypeBranch(
            "An external event that is mapped to a st2 input. It is the "
            "st2 invocation point.",
            self,
            self.subparsers,
        )

        self.commands["trigger-instance"] = triggerinstance.TriggerInstanceBranch(
            "Actual instances of triggers received by st2.", self, self.subparsers
        )

        self.commands["rule-enforcement"] = rule_enforcement.RuleEnforcementBranch(
            "Models that represent enforcement of rules.", self, self.subparsers
        )

        self.commands["workflow"] = workflow.WorkflowBranch(
            "Commands for workflow authoring related operations. "
            "Only orquesta workflows are supported.",
            self,
            self.subparsers,
        )

        # Service Registry
        self.commands["service-registry"] = service_registry.ServiceRegistryBranch(
            "Service registry group and membership related commands.",
            self,
            self.subparsers,
        )

        # RBAC
        self.commands["role"] = rbac.RoleBranch("RBAC roles.", self, self.subparsers)
        self.commands["role-assignment"] = rbac.RoleAssignmentBranch(
            "RBAC role assignments.", self, self.subparsers
        )

    def run(self, argv):
        debug = False

        parser = self.parser

        if len(argv) == 0:
            # Print a more user-friendly help string if no arguments are provided
            # Note: We only set usage variable for the main parser. If we passed "usage" argument
            # to the main ArgumentParser class above, this would also set a custom usage string for
            # sub-parsers which we don't want.
            parser.usage = USAGE_STRING
            sys.stderr.write(parser.format_help())
            return 2

        # Provide autocomplete for shell
        argcomplete.autocomplete(self.parser)

        if "--print-config" in argv:
            # Hack because --print-config requires no command to be specified
            argv = argv + ["action", "list"]

        # Parse command line arguments.
        args = self.parser.parse_args(args=argv)

        print_config = args.print_config
        if print_config:
            self._print_config(args=args)
            return 3

        # Parse config and store it in the config module
        config = self._parse_config_file(args=args, validate_config_permissions=False)
        set_config(config=config)

        self._check_locale_and_print_warning()

        # Setup client and run the command
        try:
            debug = getattr(args, "debug", False)
            if debug:
                set_log_level_for_all_loggers(level=logging.DEBUG)

            # Set up client.
            self.client = self.get_client(args=args, debug=debug)

            # TODO: This is not so nice work-around for Python 3 because of a breaking change in
            # Python 3 - https://bugs.python.org/issue16308
            try:
                func = getattr(args, "func")
            except AttributeError:
                parser.print_help()
                sys.exit(2)

            # Execute command.
            func(args)

            return 0
        except OperationFailureException:
            if debug:
                self._print_debug_info(args=args)
            return 2
        except Exception as e:
            # We allow exception to define custom exit codes
            exit_code = getattr(e, "exit_code", 1)

            print("ERROR: %s\n" % e)
            if debug:
                self._print_debug_info(args=args)

            return exit_code

    def _print_config(self, args):
        config = self._parse_config_file(args=args)

        for section, options in six.iteritems(config):
            print("[%s]" % (section))

            for name, value in six.iteritems(options):
                print("%s = %s" % (name, value))

    def _check_locale_and_print_warning(self):
        """
        Method which checks that unicode locale is used and prints a warning if it's not.
        """
        try:
            default_locale = locale.getdefaultlocale()[0]
            preferred_encoding = locale.getpreferredencoding()
        except ValueError:
            # Ignore unknown locale errors for now
            default_locale = "unknown"
            preferred_encoding = "unknown"

        if preferred_encoding and preferred_encoding.lower() != "utf-8":
            msg = NON_UTF8_LOCALE % (default_locale or "unknown", preferred_encoding)
            LOGGER.warn(msg)


def setup_logging(argv):
    debug = "--debug" in argv

    root = LOGGER
    root.setLevel(logging.WARNING)

    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.WARNING)
    formatter = logging.Formatter("%(asctime)s  %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    if not debug:
        handler.addFilter(LogLevelFilter(log_levels=[logging.ERROR]))

    root.addHandler(handler)


def main(argv=sys.argv[1:]):
    setup_logging(argv)

    if "--enable-profiler" in sys.argv:
        setup_regular_profiler(service_name="st2cli")
        sys.argv.remove("--enable-profiler")
        argv.remove("--enable-profiler")

    return Shell().run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import logging
import mock
import os
import tempfile

from st2client.commands import workflow
from st2client import shell
from st2client import models
from st2client.utils import httpclient
from tests import base as st2cli_tests


LOG = logging.getLogger(__name__)

MOCK_ACTION = {
    "ref": "mock.foobar",
    "runner_type": "mock-runner",
    "pack": "mock",
    "name": "foobar",
    "parameters": {},
    "enabled": True,
    "entry_point": "workflows/foobar.yaml",
}

MOCK_WF_DEF = """
version: 1.0
description: A basic sequential workflow.
tasks:
  task1:
    action: core.echo message="Hello, World!"
"""

MOCK_RESULT = []


def get_by_ref(**kwargs):
    return models.Action(**MOCK_ACTION)


class WorkflowCommandTestCase(st2cli_tests.BaseCLITestCase):
    def __init__(self, *args, **kwargs):
        super(WorkflowCommandTestCase, self).__init__(*args, **kwargs)
        self.shell = shell.Shell()

    @mock.patch.object(
        httpclient.HTTPClient,
        "post_raw",
        mock.MagicMock(
            return_value=st2cli_tests.FakeResponse(json.dumps(MOCK_RESULT), 200, "OK")
        ),
    )
    def test_inspect_file(self):
        fd, path = tempfile.mkstemp(suffix=".yaml")

        try:
            with open(path, "a") as f:
                f.write(MOCK_WF_DEF)

            retcode = self.shell.run(["workflow", "inspect", "--file", path])

            self.assertEqual(retcode, 0)

            httpclient.HTTPClient.post_raw.assert_called_with(
                "/inspect", MOCK_WF_DEF, headers={"content-type": "text/plain"}
            )
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        httpclient.HTTPClient,
        "post_raw",
        mock.MagicMock(
            return_value=st2cli_tests.FakeResponse(json.dumps(MOCK_RESULT), 200, "OK")
        ),
    )
    def test_inspect_bad_file(self):
        retcode = self.shell.run(["workflow", "inspect", "--file", "/tmp/foobar"])

        self.assertEqual(retcode, 1)
        self.assertIn("does not exist", self.stdout.getvalue())
        self.assertFalse(httpclient.HTTPClient.post_raw.called)

    @mock.patch.object(
        models.ResourceManager,
        "get_by_ref_or_id",
        mock.MagicMock(side_effect=get_by_ref),
    )
    @mock.patch.object(
        workflow.WorkflowInspectionCommand,
        "get_file_content",
        mock.MagicMock(return_value=MOCK_WF_DEF),
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post_raw",
        mock.MagicMock(
            return_value=st2cli_tests.FakeResponse(json.dumps(MOCK_RESULT), 200, "OK")
        ),
    )
    def test_inspect_action(self):
        retcode = self.shell.run(["workflow", "inspect", "--action", "mock.foobar"])

        self.assertEqual(retcode, 0)

        httpclient.HTTPClient.post_raw.assert_called_with(
            "/inspect", MOCK_WF_DEF, headers={"content-type": "text/plain"}
        )

    @mock.patch.object(
        models.ResourceManager, "get_by_ref_or_id", mock.MagicMock(return_value=None)
    )
    @mock.patch.object(
        httpclient.HTTPClient,
        "post_raw",
        mock.MagicMock(
            return_value=st2cli_tests.FakeResponse(json.dumps(MOCK_RESULT), 200, "OK")
        ),
    )
    def test_inspect_bad_action(self):
        retcode = self.shell.run(["workflow", "inspect", "--action", "mock.foobar"])

        self.assertEqual(retcode, 1)
        self.assertIn("Unable to identify action", self.stdout.getvalue())
        self.assertFalse(httpclient.HTTPClient.post_raw.called)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import json
import math
import logging
import sys

import six
from prettytable import PrettyTable
from six.moves import zip
from six.moves import range

from st2client import formatters
from st2client.utils import strutil
from st2client.utils.terminal import get_terminal_size_columns


LOG = logging.getLogger(__name__)

# Minimum width for the ID to make sure the ID column doesn't wrap across
# multiple lines
MIN_ID_COL_WIDTH = 26

# Minimum width for a column
MIN_COL_WIDTH = 5

# Default attribute display order to use if one is not provided
DEFAULT_ATTRIBUTE_DISPLAY_ORDER = ["id", "name", "pack", "description"]

# Attributes which contain bash escape sequences - we can't split those across multiple lines
# since things would break
COLORIZED_ATTRIBUTES = {
    "status": {"col_width": 24}  # Note: len('succeed' + ' (XXXX elapsed)') <= 24
}


class MultiColumnTable(formatters.Formatter):
    def __init__(self):
        self._table_width = 0

    @classmethod
    def format(cls, entries, *args, **kwargs):
        attributes = kwargs.get("attributes", [])
        attribute_transform_functions = kwargs.get("attribute_transform_functions", {})
        widths = kwargs.get("widths", [])
        widths = widths or []

        if not widths and attributes:
            # Dynamically calculate column size based on the terminal size
            cols = get_terminal_size_columns()

            if attributes[0] == "id":
                # consume iterator and save as entries so collection is accessible later.
                entries = [e for e in entries]
                # first column contains id, make sure it's not broken up
                first_col_width = cls._get_required_column_width(
                    values=[e.id for e in entries], minimum_width=MIN_ID_COL_WIDTH
                )
                cols = cols - first_col_width
                col_width = int(math.floor((cols / len(attributes))))
            else:
                col_width = int(math.floor((cols / len(attributes))))
                first_col_width = col_width
            widths = []
            subtract = 0
            for index in range(0, len(attributes)):
                attribute_name = attributes[index]

                if index == 0:
                    widths.append(first_col_width)
                    continue

                if attribute_name in COLORIZED_ATTRIBUTES:
                    current_col_width = COLORIZED_ATTRIBUTES[attribute_name][
                        "col_width"
                    ]
                    subtract += current_col_width - col_width
                else:
                    # Make sure we subtract the added width from the last column so we account
                    # for the fixed width columns and make sure table is not wider than the
                    # terminal width.
                    if index == (len(attributes) - 1) and subtract:
                        current_col_width = col_width - subtract

                        if current_col_width <= MIN_COL_WIDTH:
                            # Make sure column width is always grater than MIN_COL_WIDTH
                            current_col_width = MIN_COL_WIDTH
                    else:
                        current_col_width = col_width

                widths.append(current_col_width)

        if not attributes or "all" in attributes:
            entries = list(entries) if entries else []

            if len(entries) >= 1:
                attributes = list(entries[0].__dict__.keys())
                attributes = sorted(
                    [attr for attr in attributes if not attr.startswith("_")]
                )
            else:
                # There are no entries so we can't infer available attributes
                attributes = []

        # Determine table format.
        if len(attributes) == len(widths):
            # Customize width for each column.
            columns = zip(attributes, widths)
        else:
            # If only 1 width value is provided then
            # apply it to all columns else fix at 28.
            width = widths[0] if len(widths) == 1 else 28
            columns = zip(attributes, [width for i in range(0, len(attributes))])

        # Format result to table.
        table = PrettyTable()
        for column in columns:
            table.field_names.append(column[0])
            table.max_width[column[0]] = column[1]
        table.padding_width = 1
        table.align = "l"
        table.valign = "t"
        for entry in entries:
            # TODO: Improve getting values of nested dict.
            values = []
            for field_name in table.field_names:
                if "." in field_name:
                    field_names = field_name.split(".")
                    value = getattr(entry, field_names.pop(0), {})
                    for name in field_names:
                        value = cls._get_field_value(value, name)
                        if type(value) is str:
                            break
                    value = strutil.strip_carriage_returns(strutil.unescape(value))
                    values.append(value)
                else:
                    value = cls._get_simple_field_value(entry, field_name)
                    transform_function = attribute_transform_functions.get(
                        field_name, lambda value: value
                    )
                    value = transform_function(value=value)
                    value = strutil.strip_carriage_returns(strutil.unescape(value))
                    values.append(value)
            table.add_row(values)

        # width for the note
        try:
            cls.table_width = len(table.get_string().split("\n")[0])
        except IndexError:
            cls.table_width = 0

        return table

    @property
    def table_width(self):
        return self._table_width

    @table_width.setter
    def table_width(self, value):
        self._table_width = value

    @staticmethod
    def _get_simple_field_value(entry, field_name):
        """
        Format a value for a simple field.
        """
        value = getattr(entry, field_name, "")
        if isinstance(value, (list, tuple)):
            if len(value) == 0:
                value = ""
            elif isinstance(value[0], (str, six.text_type)):
                # List contains simple string values, format it as comma
                # separated string
                value = ", ".join(value)

        return value

    @staticmethod
    def _get_field_value(value, field_name):
        r_val = value.get(field_name, None)
        if r_val is None:
            return ""

        if isinstance(r_val, list) or isinstance(r_val, dict):
            return r_val if len(r_val) > 0 else ""
        return r_val

    @staticmethod
    def _get_friendly_column_name(name):
        if not name:
            return None

        friendly_name = name.replace("_", " ").replace(".", " ").capitalize()
        return friendly_name

    @staticmethod
    def _get_required_column_width(values, minimum_width=0):
        max_width = len(max(values, key=len)) if values else minimum_width
        return max_width if max_width > minimum_width else minimum_width


class PropertyValueTable(formatters.Formatter):
    @classmethod
    def format(cls, subject, *args, **kwargs):
        attributes = kwargs.get("attributes", None)
        attribute_display_order = kwargs.get(
            "attribute_display_order", DEFAULT_ATTRIBUTE_DISPLAY_ORDER
        )
        attribute_transform_functions = kwargs.get("attribute_transform_functions", {})

        if not attributes or "all" in attributes:
            attributes = sorted(
                [attr for attr in subject.__dict__ if not attr.startswith("_")]
            )

        for attr in attribute_display_order[::-1]:
            if attr in attributes:
                attributes.remove(attr)
                attributes = [attr] + attributes
        table = PrettyTable()
        table.field_names = ["Property", "Value"]
        table.max_width["Property"] = 20
        table.max_width["Value"] = 60
        table.padding_width = 1
        table.align = "l"
        table.valign = "t"

        for attribute in attributes:
            if "." in attribute:
                field_names = attribute.split(".")
                value = cls._get_attribute_value(subject, field_names.pop(0))
                for name in field_names:
                    value = cls._get_attribute_value(value, name)
                    if type(value) is str:
                        break
            else:
                value = cls._get_attribute_value(subject, attribute)

            transform_function = attribute_transform_functions.get(
                attribute, lambda value: value
            )
            value = transform_function(value=value)

            if type(value) is dict or type(value) is list:
                value = json.dumps(value, indent=4)

            value = strutil.strip_carriage_returns(strutil.unescape(value))
            table.add_row([attribute, value])
        return table

    @staticmethod
    def _get_attribute_value(subject, attribute):
        if isinstance(subject, dict):
            r_val = subject.get(attribute, None)
        else:
            r_val = getattr(subject, attribute, None)
        if r_val is None:
            return ""
        if isinstance(r_val, list) or isinstance(r_val, dict):
            return r_val if len(r_val) > 0 else ""
        return r_val


class SingleRowTable(object):
    @staticmethod
    def note_box(entity, limit):
        if limit <= 0:
            return None
        elif limit == 1:

            if entity == "inquiries":
                entity = "inquiry"
            else:
                entity = entity[:-1]

            message = (
                "Note: Only one %s is displayed. Use -n/--last flag for more results."
                % entity
            )
        else:
            message = (
                "Note: Only first %s %s are displayed. Use -n/--last flag for more results."
                % (limit, entity)
            )
        # adding default padding
        message_length = len(message) + 3
        m = MultiColumnTable()
        if m.table_width > message_length:
            note = PrettyTable(
                [""], right_padding_width=(m.table_width - message_length)
            )
        else:
            note = PrettyTable([""])
        note.header = False
        note.add_row([message])
        sys.stderr.write((str(note) + "\n"))
        return



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os
import uuid
import json
import mock
import tempfile
import requests
import argparse
import logging

import six

from tests import base
from st2client import shell
from st2client.models.core import add_auth_token_to_kwargs_from_env
from st2client.commands.resource import add_auth_token_to_kwargs_from_cli
from st2client.utils.httpclient import (
    add_auth_token_to_headers,
    add_json_content_type_to_headers,
)


LOG = logging.getLogger(__name__)

if six.PY3:
    RULE = {
        "name": "drule",
        "description": "i am THE rule.",
        "pack": "cli",
        "id": uuid.uuid4().hex,
    }
else:
    RULE = {
        "id": uuid.uuid4().hex,
        "description": "i am THE rule.",
        "name": "drule",
        "pack": "cli",
    }


class TestLoginBase(base.BaseCLITestCase):
    """
    A base class for testing related to 'st2 login' commands

    This exists primarily to ensure that each specific test case is kept atomic,
    since the tests create actual files on the filesystem - as well as to cut down
    on duplicate code in each test class
    """

    DOTST2_PATH = os.path.expanduser("~/.st2/")
    CONFIG_FILE_NAME = "st2.conf"
    PARENT_DIR = "testconfig"
    TMP_DIR = tempfile.mkdtemp()
    CONFIG_CONTENTS = """
    [credentials]
    username = olduser
    password = Password1!
    """

    def __init__(self, *args, **kwargs):
        super(TestLoginBase, self).__init__(*args, **kwargs)

        # We're overriding the default behavior for CLI test cases here
        self.DEFAULT_SKIP_CONFIG = "0"

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("-t", "--token", dest="token")
        self.parser.add_argument("--api-key", dest="api_key")
        self.shell = shell.Shell()

        self.CONFIG_DIR = os.path.join(self.TMP_DIR, self.PARENT_DIR)
        self.CONFIG_FILE = os.path.join(self.CONFIG_DIR, self.CONFIG_FILE_NAME)

    def setUp(self):
        super(TestLoginBase, self).setUp()

        if not os.path.isdir(self.CONFIG_DIR):
            os.makedirs(self.CONFIG_DIR)
            os.chmod(self.CONFIG_DIR, 0o2770)

        # Remove any existing config file
        if os.path.isfile(self.CONFIG_FILE):
            os.remove(self.CONFIG_FILE)

        with open(self.CONFIG_FILE, "w") as cfg:
            for line in self.CONFIG_CONTENTS.split("\n"):
                cfg.write("%s\n" % line.strip())

        os.chmod(self.CONFIG_FILE, 0o660)

    def tearDown(self):
        super(TestLoginBase, self).tearDown()

        # Clean up config file
        os.remove(self.CONFIG_FILE)

        # Clean up tokens
        for file in [f for f in os.listdir(self.DOTST2_PATH) if "token-" in f]:
            os.remove(self.DOTST2_PATH + file)

        # Clean up config directory
        os.rmdir(self.CONFIG_DIR)


class TestLoginPasswordAndConfig(TestLoginBase):

    CONFIG_FILE_NAME = "logintest.cfg"

    TOKEN = {
        "user": "st2admin",
        "token": "44583f15945b4095afbf57058535ca64",
        "expiry": "2017-02-12T00:53:09.632783Z",
        "id": "589e607532ed3535707f10eb",
        "metadata": {},
    }

    @mock.patch.object(
        requests,
        "post",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(TOKEN), 200, "OK")),
    )
    def runTest(self):
        """Test 'st2 login' functionality by specifying a password and a configuration file"""

        expected_username = self.TOKEN["user"]
        args = [
            "--config",
            self.CONFIG_FILE,
            "login",
            expected_username,
            "--password",
            "Password1!",
        ]

        self.shell.run(args)

        with open(self.CONFIG_FILE, "r") as config_file:
            for line in config_file.readlines():
                # Make sure certain values are not present
                self.assertNotIn("password", line)
                self.assertNotIn("olduser", line)

                # Make sure configured username is what we expect
                if "username" in line:
                    self.assertEqual(line.split(" ")[2][:-1], expected_username)

            # validate token was created
            self.assertTrue(
                os.path.isfile("%stoken-%s" % (self.DOTST2_PATH, expected_username))
            )


class TestLoginIntPwdAndConfig(TestLoginBase):

    CONFIG_FILE_NAME = "logintest.cfg"

    TOKEN = {
        "user": "st2admin",
        "token": "44583f15945b4095afbf57058535ca64",
        "expiry": "2017-02-12T00:53:09.632783Z",
        "id": "589e607532ed3535707f10eb",
        "metadata": {},
    }

    @mock.patch.object(
        requests,
        "post",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(TOKEN), 200, "OK")),
    )
    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps({}), 200, "OK")),
    )
    @mock.patch("st2client.commands.auth.getpass")
    def runTest(self, mock_gp):
        """Test 'st2 login' functionality with interactive password entry"""

        expected_username = self.TOKEN["user"]
        args = ["--config", self.CONFIG_FILE, "login", expected_username]

        mock_gp.getpass.return_value = "Password1!"

        self.shell.run(args)

        expected_kwargs = {
            "headers": {"content-type": "application/json"},
            "auth": ("st2admin", "Password1!"),
        }
        requests.post.assert_called_with(
            "http://127.0.0.1:9100/tokens", "{}", **expected_kwargs
        )

        # Check file permissions
        self.assertEqual(os.stat(self.CONFIG_FILE).st_mode & 0o777, 0o660)

        with open(self.CONFIG_FILE, "r") as config_file:
            for line in config_file.readlines():
                # Make sure certain values are not present
                self.assertNotIn("password", line)
                self.assertNotIn("olduser", line)

                # Make sure configured username is what we expect
                if "username" in line:
                    self.assertEqual(line.split(" ")[2][:-1], expected_username)

            # validate token was created
            self.assertTrue(
                os.path.isfile("%stoken-%s" % (self.DOTST2_PATH, expected_username))
            )

        # Validate token is sent on subsequent requests to st2 API
        args = ["--config", self.CONFIG_FILE, "pack", "list"]
        self.shell.run(args)

        expected_kwargs = {
            "headers": {"X-Auth-Token": self.TOKEN["token"]},
            "params": {"include_attributes": "ref,name,description,version,author"},
        }
        requests.get.assert_called_with(
            "http://127.0.0.1:9101/v1/packs", **expected_kwargs
        )


class TestLoginWritePwdOkay(TestLoginBase):

    CONFIG_FILE_NAME = "logintest.cfg"

    TOKEN = {
        "user": "st2admin",
        "token": "44583f15945b4095afbf57058535ca64",
        "expiry": "2017-02-12T00:53:09.632783Z",
        "id": "589e607532ed3535707f10eb",
        "metadata": {},
    }

    @mock.patch.object(
        requests,
        "post",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(TOKEN), 200, "OK")),
    )
    @mock.patch("st2client.commands.auth.getpass")
    def runTest(self, mock_gp):
        """Test 'st2 login' functionality with --write-password flag set"""

        expected_username = self.TOKEN["user"]
        args = [
            "--config",
            self.CONFIG_FILE,
            "login",
            expected_username,
            "--password",
            "Password1!",
            "--write-password",
        ]

        self.shell.run(args)

        with open(self.CONFIG_FILE, "r") as config_file:

            for line in config_file.readlines():

                # Make sure certain values are not present
                self.assertNotIn("olduser", line)

                # Make sure configured username is what we expect
                if "username" in line:
                    self.assertEqual(line.split(" ")[2][:-1], expected_username)

            # validate token was created
            self.assertTrue(
                os.path.isfile("%stoken-%s" % (self.DOTST2_PATH, expected_username))
            )


class TestLoginUncaughtException(TestLoginBase):

    CONFIG_FILE_NAME = "logintest.cfg"

    TOKEN = {
        "user": "st2admin",
        "token": "44583f15945b4095afbf57058535ca64",
        "expiry": "2017-02-12T00:53:09.632783Z",
        "id": "589e607532ed3535707f10eb",
        "metadata": {},
    }

    @mock.patch.object(
        requests,
        "post",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(TOKEN), 200, "OK")),
    )
    @mock.patch("st2client.commands.auth.getpass")
    def runTest(self, mock_gp):
        """Test 'st2 login' ability to detect unhandled exceptions"""

        expected_username = self.TOKEN["user"]
        args = ["--config", self.CONFIG_FILE, "login", expected_username]

        mock_gp.getpass = mock.MagicMock(side_effect=Exception)

        self.shell.run(args)
        retcode = self.shell.run(args)

        self.assertIn(
            "Failed to log in as %s" % expected_username, self.stdout.getvalue()
        )
        self.assertNotIn("Logged in as", self.stdout.getvalue())
        self.assertEqual(retcode, 1)


class TestAuthToken(base.BaseCLITestCase):

    capture_output = False

    def __init__(self, *args, **kwargs):
        super(TestAuthToken, self).__init__(*args, **kwargs)
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("-t", "--token", dest="token")
        self.parser.add_argument("--api-key", dest="api_key")
        self.shell = shell.Shell()

    def setUp(self):
        super(TestAuthToken, self).setUp()

        # Setup environment.
        os.environ["ST2_BASE_URL"] = "http://127.0.0.1"

    def tearDown(self):
        super(TestAuthToken, self).tearDown()

        # Clean up environment.
        if "ST2_AUTH_TOKEN" in os.environ:
            del os.environ["ST2_AUTH_TOKEN"]
        if "ST2_API_KEY" in os.environ:
            del os.environ["ST2_API_KEY"]
        if "ST2_BASE_URL" in os.environ:
            del os.environ["ST2_BASE_URL"]

    @add_auth_token_to_kwargs_from_cli
    @add_auth_token_to_kwargs_from_env
    def _mock_run(self, args, **kwargs):
        return kwargs

    def test_decorate_auth_token_by_cli(self):
        token = uuid.uuid4().hex
        args = self.parser.parse_args(args=["-t", token])
        self.assertDictEqual(self._mock_run(args), {"token": token})
        args = self.parser.parse_args(args=["--token", token])
        self.assertDictEqual(self._mock_run(args), {"token": token})

    def test_decorate_api_key_by_cli(self):
        token = uuid.uuid4().hex
        args = self.parser.parse_args(args=["--api-key", token])
        self.assertDictEqual(self._mock_run(args), {"api_key": token})

    def test_decorate_auth_token_by_env(self):
        token = uuid.uuid4().hex
        os.environ["ST2_AUTH_TOKEN"] = token
        args = self.parser.parse_args(args=[])
        self.assertDictEqual(self._mock_run(args), {"token": token})

    def test_decorate_api_key_by_env(self):
        token = uuid.uuid4().hex
        os.environ["ST2_API_KEY"] = token
        args = self.parser.parse_args(args=[])
        self.assertDictEqual(self._mock_run(args), {"api_key": token})

    def test_decorate_without_auth_token(self):
        args = self.parser.parse_args(args=[])
        self.assertDictEqual(self._mock_run(args), {})

    @add_auth_token_to_headers
    @add_json_content_type_to_headers
    def _mock_http(self, url, **kwargs):
        return kwargs

    def test_decorate_auth_token_to_http_headers(self):
        token = uuid.uuid4().hex
        kwargs = self._mock_http("/", token=token)
        expected = {"content-type": "application/json", "X-Auth-Token": token}
        self.assertIn("headers", kwargs)
        self.assertDictEqual(kwargs["headers"], expected)

    def test_decorate_api_key_to_http_headers(self):
        token = uuid.uuid4().hex
        kwargs = self._mock_http("/", api_key=token)
        expected = {"content-type": "application/json", "St2-Api-Key": token}
        self.assertIn("headers", kwargs)
        self.assertDictEqual(kwargs["headers"], expected)

    def test_decorate_without_auth_token_to_http_headers(self):
        kwargs = self._mock_http("/", auth=("stanley", "stanley"))
        expected = {"content-type": "application/json"}
        self.assertIn("auth", kwargs)
        self.assertEqual(kwargs["auth"], ("stanley", "stanley"))
        self.assertIn("headers", kwargs)
        self.assertDictEqual(kwargs["headers"], expected)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps({}), 200, "OK")),
    )
    def test_decorate_resource_list(self):
        url = (
            "http://127.0.0.1:9101/v1/rules/"
            "?include_attributes=ref,pack,description,enabled&limit=50"
        )
        url = url.replace(",", "%2C")

        # Test without token.
        self.shell.run(["rule", "list"])
        kwargs = {}
        requests.get.assert_called_with(url, **kwargs)

        # Test with token from  cli.
        token = uuid.uuid4().hex
        self.shell.run(["rule", "list", "-t", token])
        kwargs = {"headers": {"X-Auth-Token": token}}
        requests.get.assert_called_with(url, **kwargs)

        # Test with token from env.
        token = uuid.uuid4().hex
        os.environ["ST2_AUTH_TOKEN"] = token
        self.shell.run(["rule", "list"])
        kwargs = {"headers": {"X-Auth-Token": token}}
        requests.get.assert_called_with(url, **kwargs)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(RULE), 200, "OK")),
    )
    def test_decorate_resource_get(self):
        rule_ref = "%s.%s" % (RULE["pack"], RULE["name"])
        url = "http://127.0.0.1:9101/v1/rules/%s" % rule_ref

        # Test without token.
        self.shell.run(["rule", "get", rule_ref])
        kwargs = {}
        requests.get.assert_called_with(url, **kwargs)

        # Test with token from cli.
        token = uuid.uuid4().hex
        self.shell.run(["rule", "get", rule_ref, "-t", token])
        kwargs = {"headers": {"X-Auth-Token": token}}
        requests.get.assert_called_with(url, **kwargs)

        # Test with token from env.
        token = uuid.uuid4().hex
        os.environ["ST2_AUTH_TOKEN"] = token
        self.shell.run(["rule", "get", rule_ref])
        kwargs = {"headers": {"X-Auth-Token": token}}
        requests.get.assert_called_with(url, **kwargs)

    @mock.patch.object(
        requests,
        "post",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(RULE), 200, "OK")),
    )
    def test_decorate_resource_post(self):
        url = "http://127.0.0.1:9101/v1/rules"
        data = {"name": RULE["name"], "description": RULE["description"]}

        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(data, indent=4))

            # Test without token.
            self.shell.run(["rule", "create", path])
            kwargs = {"headers": {"content-type": "application/json"}}
            requests.post.assert_called_with(url, json.dumps(data), **kwargs)

            # Test with token from cli.
            token = uuid.uuid4().hex
            self.shell.run(["rule", "create", path, "-t", token])
            kwargs = {
                "headers": {"content-type": "application/json", "X-Auth-Token": token}
            }
            requests.post.assert_called_with(url, json.dumps(data), **kwargs)

            # Test with token from env.
            token = uuid.uuid4().hex
            os.environ["ST2_AUTH_TOKEN"] = token
            self.shell.run(["rule", "create", path])
            kwargs = {
                "headers": {"content-type": "application/json", "X-Auth-Token": token}
            }
            requests.post.assert_called_with(url, json.dumps(data), **kwargs)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(RULE), 200, "OK")),
    )
    @mock.patch.object(
        requests,
        "put",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(RULE), 200, "OK")),
    )
    def test_decorate_resource_put(self):
        rule_ref = "%s.%s" % (RULE["pack"], RULE["name"])

        get_url = "http://127.0.0.1:9101/v1/rules/%s" % rule_ref
        put_url = "http://127.0.0.1:9101/v1/rules/%s" % RULE["id"]
        data = {
            "name": RULE["name"],
            "description": RULE["description"],
            "pack": RULE["pack"],
        }

        fd, path = tempfile.mkstemp(suffix=".json")
        try:
            with open(path, "a") as f:
                f.write(json.dumps(data, indent=4))

            # Test without token.
            self.shell.run(["rule", "update", rule_ref, path])
            kwargs = {}
            requests.get.assert_called_with(get_url, **kwargs)
            kwargs = {"headers": {"content-type": "application/json"}}
            requests.put.assert_called_with(put_url, json.dumps(RULE), **kwargs)

            # Test with token from cli.
            token = uuid.uuid4().hex
            self.shell.run(["rule", "update", rule_ref, path, "-t", token])
            kwargs = {"headers": {"X-Auth-Token": token}}
            requests.get.assert_called_with(get_url, **kwargs)
            kwargs = {
                "headers": {"content-type": "application/json", "X-Auth-Token": token}
            }

            requests.put.assert_called_with(put_url, json.dumps(RULE), **kwargs)

            # Test with token from env.
            token = uuid.uuid4().hex
            os.environ["ST2_AUTH_TOKEN"] = token
            self.shell.run(["rule", "update", rule_ref, path])
            kwargs = {"headers": {"X-Auth-Token": token}}
            requests.get.assert_called_with(get_url, **kwargs)

            # Note: We parse the payload because data might not be in the same
            # order as the fixture
            kwargs = {
                "headers": {"content-type": "application/json", "X-Auth-Token": token}
            }
            requests.put.assert_called_with(put_url, json.dumps(RULE), **kwargs)
        finally:
            os.close(fd)
            os.unlink(path)

    @mock.patch.object(
        requests,
        "get",
        mock.MagicMock(return_value=base.FakeResponse(json.dumps(RULE), 200, "OK")),
    )
    @mock.patch.object(
        requests,
        "delete",
        mock.MagicMock(return_value=base.FakeResponse("", 204, "OK")),
    )
    def test_decorate_resource_delete(self):
        rule_ref = "%s.%s" % (RULE["pack"], RULE["name"])
        get_url = "http://127.0.0.1:9101/v1/rules/%s" % rule_ref
        del_url = "http://127.0.0.1:9101/v1/rules/%s" % RULE["id"]

        # Test without token.
        self.shell.run(["rule", "delete", rule_ref])
        kwargs = {}
        requests.get.assert_called_with(get_url, **kwargs)
        requests.delete.assert_called_with(del_url, **kwargs)

        # Test with token from cli.
        token = uuid.uuid4().hex
        self.shell.run(["rule", "delete", rule_ref, "-t", token])
        kwargs = {"headers": {"X-Auth-Token": token}}
        requests.get.assert_called_with(get_url, **kwargs)
        requests.delete.assert_called_with(del_url, **kwargs)

        # Test with token from env.
        token = uuid.uuid4().hex
        os.environ["ST2_AUTH_TOKEN"] = token
        self.shell.run(["rule", "delete", rule_ref])
        kwargs = {"headers": {"X-Auth-Token": token}}
        requests.get.assert_called_with(get_url, **kwargs)
        requests.delete.assert_called_with(del_url, **kwargs)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from st2client.formatters import table
from st2client.commands import resource
from st2client.models.rbac import Role
from st2client.models.rbac import UserRoleAssignment

__all__ = ["RoleBranch", "RoleAssignmentBranch"]

ROLE_ATTRIBUTE_DISPLAY_ORDER = ["id", "name", "system", "permission_grants"]
ROLE_ASSIGNMENT_ATTRIBUTE_DISPLAY_ORDER = [
    "id",
    "role",
    "user",
    "is_remote",
    "description",
]


class RoleBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(RoleBranch, self).__init__(
            Role,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            read_only=True,
            commands={"list": RoleListCommand, "get": RoleGetCommand},
        )


class RoleListCommand(resource.ResourceCommand):
    display_attributes = ["id", "name", "system", "description"]
    attribute_display_order = ROLE_ATTRIBUTE_DISPLAY_ORDER

    def __init__(self, resource, *args, **kwargs):
        super(RoleListCommand, self).__init__(
            resource,
            "list",
            "Get the list of the  %s." % resource.get_plural_display_name().lower(),
            *args,
            **kwargs,
        )

        self.group = self.parser.add_mutually_exclusive_group()

        # Filter options
        self.group.add_argument(
            "-s", "--system", action="store_true", help="Only display system roles."
        )

        # Display options
        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" will return all '
                "attributes."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        # Filtering options
        if args.system:
            kwargs["system"] = args.system

        if args.system:
            result = self.manager.query(**kwargs)
        else:
            result = self.manager.get_all(**kwargs)

        return result

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)
        self.print_output(
            instances,
            table.MultiColumnTable,
            attributes=args.attr,
            widths=args.width,
            json=args.json,
            yaml=args.yaml,
        )


class RoleGetCommand(resource.ResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = ROLE_ATTRIBUTE_DISPLAY_ORDER
    pk_argument_name = "id"


class RoleAssignmentBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(RoleAssignmentBranch, self).__init__(
            UserRoleAssignment,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            read_only=True,
            commands={
                "list": RoleAssignmentListCommand,
                "get": RoleAssignmentGetCommand,
            },
        )


class RoleAssignmentListCommand(resource.ResourceCommand):
    display_attributes = ["id", "role", "user", "is_remote", "source", "description"]
    attribute_display_order = ROLE_ASSIGNMENT_ATTRIBUTE_DISPLAY_ORDER

    def __init__(self, resource, *args, **kwargs):
        super(RoleAssignmentListCommand, self).__init__(
            resource,
            "list",
            "Get the list of the  %s." % resource.get_plural_display_name().lower(),
            *args,
            **kwargs,
        )

        # Filter options
        self.parser.add_argument("-r", "--role", help="Role to filter on.")
        self.parser.add_argument("-u", "--user", help="User to filter on.")
        self.parser.add_argument("-s", "--source", help="Source to filter on.")
        self.parser.add_argument(
            "--remote",
            action="store_true",
            help="Only display remote role assignments.",
        )

        # Display options
        self.parser.add_argument(
            "-a",
            "--attr",
            nargs="+",
            default=self.display_attributes,
            help=(
                "List of attributes to include in the "
                'output. "all" will return all '
                "attributes."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--width",
            nargs="+",
            type=int,
            default=None,
            help=("Set the width of columns in output."),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        # Filtering options
        if args.role:
            kwargs["role"] = args.role
        if args.user:
            kwargs["user"] = args.user
        if args.source:
            kwargs["source"] = args.source
        if args.remote:
            kwargs["remote"] = args.remote

        if args.role or args.user or args.remote or args.source:
            result = self.manager.query(**kwargs)
        else:
            result = self.manager.get_all(**kwargs)

        return result

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)
        self.print_output(
            instances,
            table.MultiColumnTable,
            attributes=args.attr,
            widths=args.width,
            json=args.json,
            yaml=args.yaml,
        )


class RoleAssignmentGetCommand(resource.ResourceGetCommand):
    display_attributes = ["all"]
    attribute_display_order = ROLE_ASSIGNMENT_ATTRIBUTE_DISPLAY_ORDER
    pk_argument_name = "id"



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import logging
import mock
import re
import unittest2

import prompt_toolkit
from prompt_toolkit.document import Document
from six.moves import StringIO

from st2client.utils import interactive
import six


LOG = logging.getLogger(__name__)


class TestInteractive(unittest2.TestCase):
    def assertPromptMessage(self, prompt_mock, message, msg=None):
        self.assertEqual(prompt_mock.call_args[0], (message,), msg)

    def assertPromptDescription(self, prompt_mock, message, msg=None):
        toolbar_factory = prompt_mock.call_args[1]["get_bottom_toolbar_tokens"]
        self.assertEqual(toolbar_factory(None)[0][1], message, msg)

    def assertPromptValidate(self, prompt_mock, value):
        validator = prompt_mock.call_args[1]["validator"]

        validator.validate(Document(text=six.text_type(value)))

    def assertPromptPassword(self, prompt_mock, value, msg=None):
        self.assertEqual(prompt_mock.call_args[1]["is_password"], value, msg)

    def test_interactive_form(self):
        reader = mock.MagicMock()
        Reader = mock.MagicMock(return_value=reader)
        Reader.condition = mock.MagicMock(return_value=True)

        schema = {"string": {"type": "string"}}

        with mock.patch.object(interactive.InteractiveForm, "readers", [Reader]):
            interactive.InteractiveForm(schema).initiate_dialog()

        Reader.condition.assert_called_once_with(schema["string"])
        reader.read.assert_called_once_with()

    def test_interactive_form_no_match(self):
        reader = mock.MagicMock()
        Reader = mock.MagicMock(return_value=reader)
        Reader.condition = mock.MagicMock(return_value=False)

        schema = {"string": {"type": "string"}}

        with mock.patch.object(interactive.InteractiveForm, "readers", [Reader]):
            interactive.InteractiveForm(schema).initiate_dialog()

        Reader.condition.assert_called_once_with(schema["string"])
        reader.read.assert_not_called()

    @mock.patch("sys.stdout", new_callable=StringIO)
    def test_interactive_form_interrupted(self, stdout_mock):
        reader = mock.MagicMock()
        Reader = mock.MagicMock(return_value=reader)
        Reader.condition = mock.MagicMock(return_value=True)
        reader.read = mock.MagicMock(side_effect=KeyboardInterrupt)

        schema = {"string": {"type": "string"}}

        with mock.patch.object(interactive.InteractiveForm, "readers", [Reader]):
            interactive.InteractiveForm(schema).initiate_dialog()

        self.assertEqual(stdout_mock.getvalue(), "Dialog interrupted.\n")

    def test_interactive_form_interrupted_reraised(self):
        reader = mock.MagicMock()
        Reader = mock.MagicMock(return_value=reader)
        Reader.condition = mock.MagicMock(return_value=True)
        reader.read = mock.MagicMock(side_effect=KeyboardInterrupt)

        schema = {"string": {"type": "string"}}

        with mock.patch.object(interactive.InteractiveForm, "readers", [Reader]):
            self.assertRaises(
                interactive.DialogInterrupted,
                interactive.InteractiveForm(schema, reraise=True).initiate_dialog,
            )

    @mock.patch.object(interactive, "prompt")
    def test_stringreader(self, prompt_mock):
        spec = {"description": "some description", "default": "hey"}
        Reader = interactive.StringReader("some", spec)

        prompt_mock.return_value = "stuff"
        result = Reader.read()

        self.assertEqual(result, "stuff")
        self.assertPromptMessage(prompt_mock, "some [hey]: ")
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "stuff")

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, "hey")
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_booleanreader(self, prompt_mock):
        spec = {"description": "some description", "default": False}
        Reader = interactive.BooleanReader("some", spec)

        prompt_mock.return_value = "y"
        result = Reader.read()

        self.assertEqual(result, True)
        self.assertPromptMessage(prompt_mock, "some (boolean) [n]: ")
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "y")
        self.assertRaises(
            prompt_toolkit.validation.ValidationError,
            self.assertPromptValidate,
            prompt_mock,
            "some",
        )

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, False)
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_numberreader(self, prompt_mock):
        spec = {"description": "some description", "default": 3.2}
        Reader = interactive.NumberReader("some", spec)

        prompt_mock.return_value = "5.3"
        result = Reader.read()

        self.assertEqual(result, 5.3)
        self.assertPromptMessage(prompt_mock, "some (float) [3.2]: ")
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "5.3")
        self.assertRaises(
            prompt_toolkit.validation.ValidationError,
            self.assertPromptValidate,
            prompt_mock,
            "some",
        )

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, 3.2)
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_integerreader(self, prompt_mock):
        spec = {"description": "some description", "default": 3}
        Reader = interactive.IntegerReader("some", spec)

        prompt_mock.return_value = "5"
        result = Reader.read()

        self.assertEqual(result, 5)
        self.assertPromptMessage(prompt_mock, "some (integer) [3]: ")
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "5")
        self.assertRaises(
            prompt_toolkit.validation.ValidationError,
            self.assertPromptValidate,
            prompt_mock,
            "5.3",
        )

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, 3)
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_secretstringreader(self, prompt_mock):
        spec = {"description": "some description", "default": "hey"}
        Reader = interactive.SecretStringReader("some", spec)

        prompt_mock.return_value = "stuff"
        result = Reader.read()

        self.assertEqual(result, "stuff")
        self.assertPromptMessage(prompt_mock, "some (secret) [hey]: ")
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "stuff")
        self.assertPromptPassword(prompt_mock, True)

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, "hey")
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_enumreader(self, prompt_mock):
        spec = {
            "enum": ["some", "thing", "else"],
            "description": "some description",
            "default": "thing",
        }
        Reader = interactive.EnumReader("some", spec)

        prompt_mock.return_value = "2"
        result = Reader.read()

        self.assertEqual(result, "else")
        message = "some: \n 0 - some\n 1 - thing\n 2 - else\nChoose from 0, 1, 2 [1]: "
        self.assertPromptMessage(prompt_mock, message)
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "0")
        self.assertRaises(
            prompt_toolkit.validation.ValidationError,
            self.assertPromptValidate,
            prompt_mock,
            "some",
        )
        self.assertRaises(
            prompt_toolkit.validation.ValidationError,
            self.assertPromptValidate,
            prompt_mock,
            "5",
        )

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, "thing")
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_arrayreader(self, prompt_mock):
        spec = {"description": "some description", "default": ["a", "b"]}
        Reader = interactive.ArrayReader("some", spec)

        prompt_mock.return_value = "some,thing,else"
        result = Reader.read()

        self.assertEqual(result, ["some", "thing", "else"])
        self.assertPromptMessage(prompt_mock, "some (comma-separated list) [a,b]: ")
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "some,thing,else")

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, ["a", "b"])
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_arrayreader_ends_with_comma(self, prompt_mock):
        spec = {"description": "some description", "default": ["a", "b"]}
        Reader = interactive.ArrayReader("some", spec)

        prompt_mock.return_value = "some,thing,else,"
        result = Reader.read()

        self.assertEqual(result, ["some", "thing", "else", ""])
        self.assertPromptMessage(prompt_mock, "some (comma-separated list) [a,b]: ")
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "some,thing,else,")

    @mock.patch.object(interactive, "prompt")
    def test_arrayenumreader(self, prompt_mock):
        spec = {
            "items": {"enum": ["a", "b", "c", "d", "e"]},
            "description": "some description",
            "default": ["a", "b"],
        }
        Reader = interactive.ArrayEnumReader("some", spec)

        prompt_mock.return_value = "0,2,4"
        result = Reader.read()

        self.assertEqual(result, ["a", "c", "e"])
        message = "some: \n 0 - a\n 1 - b\n 2 - c\n 3 - d\n 4 - e\nChoose from 0, 1, 2... [0, 1]: "
        self.assertPromptMessage(prompt_mock, message)
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "0,2,4")

        prompt_mock.return_value = ""
        result = Reader.read()

        self.assertEqual(result, ["a", "b"])
        self.assertPromptValidate(prompt_mock, "")

    @mock.patch.object(interactive, "prompt")
    def test_arrayenumreader_ends_with_comma(self, prompt_mock):
        spec = {
            "items": {"enum": ["a", "b", "c", "d", "e"]},
            "description": "some description",
            "default": ["a", "b"],
        }
        Reader = interactive.ArrayEnumReader("some", spec)

        prompt_mock.return_value = "0,2,4,"
        result = Reader.read()

        self.assertEqual(result, ["a", "c", "e"])
        message = "some: \n 0 - a\n 1 - b\n 2 - c\n 3 - d\n 4 - e\nChoose from 0, 1, 2... [0, 1]: "
        self.assertPromptMessage(prompt_mock, message)
        self.assertPromptDescription(prompt_mock, "some description")
        self.assertPromptValidate(prompt_mock, "0,2,4,")

    @mock.patch.object(interactive, "prompt")
    def test_arrayobjectreader(self, prompt_mock):
        spec = {
            "items": {
                "type": "object",
                "properties": {
                    "foo": {
                        "type": "string",
                        "description": "some description",
                    },
                    "bar": {
                        "type": "string",
                        "description": "some description",
                    },
                },
            },
            "description": "some description",
        }
        Reader = interactive.ArrayObjectReader("some", spec)

        # To emulate continuing setting, this flag variable is needed
        self.is_continued = False

        def side_effect(msg, **kwargs):
            if re.match(r"^~~~ Would you like to add another item to.*", msg):
                # prompt requires the input to judge continuing setting, or not
                if not self.is_continued:
                    # continuing the configuration only once
                    self.is_continued = True
                    return ""
                else:
                    # finishing to configuration
                    return "n"
            else:
                # prompt requires the input of property value in the object
                return "value"

        prompt_mock.side_effect = side_effect
        results = Reader.read()

        self.assertEqual(len(results), 2)
        self.assertTrue(all([len(list(x.keys())) == 2 for x in results]))
        self.assertTrue(all(["foo" in x and "bar" in x for x in results]))



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from st2client.models import core


LOG = logging.getLogger(__name__)


class RunnerType(core.Resource):
    _alias = "Runner"
    _display_name = "Runner"
    _plural = "RunnerTypes"
    _plural_display_name = "Runners"
    _repr_attributes = ["name", "enabled", "description"]


class Action(core.Resource):
    _plural = "Actions"
    _repr_attributes = ["name", "pack", "enabled", "runner_type"]
    _url_path = "actions"


class Execution(core.Resource):
    _alias = "Execution"
    _display_name = "Action Execution"
    _url_path = "executions"
    _plural = "ActionExecutions"
    _plural_display_name = "Action executions"
    _repr_attributes = [
        "status",
        "action",
        "start_timestamp",
        "end_timestamp",
        "parameters",
        "delay",
    ]


# NOTE: LiveAction has been deprecated in favor of Execution. It will be left here for
# backward compatibility reasons until v3.2.0
class LiveAction(Execution):
    pass



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import os
import yaml

from st2client import commands
from st2client.commands.resource import add_auth_token_to_kwargs_from_cli


LOG = logging.getLogger(__name__)


class WorkflowBranch(commands.Branch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(WorkflowBranch, self).__init__(
            "workflow", description, app, subparsers, parent_parser=parent_parser
        )

        # Add subparser to register subcommands for managing workflows.
        help_message = "List of commands for managing workflows."
        self.subparsers = self.parser.add_subparsers(help=help_message)

        # Register workflow commands.
        self.commands["inspect"] = WorkflowInspectionCommand(self.app, self.subparsers)


class WorkflowInspectionCommand(commands.Command):
    def __init__(self, *args, **kwargs):
        name = "inspect"
        description = (
            "Inspect workflow definition and return the list of errors if any."
        )
        args = tuple([name, description] + list(args))
        super(WorkflowInspectionCommand, self).__init__(*args, **kwargs)

        # Add argument options.
        arg_group = self.parser.add_mutually_exclusive_group()

        arg_group.add_argument(
            "--file", dest="file", help="Local file path to the workflow definition."
        )

        arg_group.add_argument(
            "--action",
            dest="action",
            help="Reference name for the registered action. This option works only if the file "
            "referenced by the entry point is installed locally under /opt/stackstorm/packs.",
        )

    @property
    def manager(self):
        return self.app.client.managers["Workflow"]

    def get_file_content(self, file_path):
        if not os.path.isfile(file_path):
            raise Exception('File "%s" does not exist on local system.' % file_path)

        with open(file_path, "r") as f:
            content = f.read()

        return content

    @add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        wf_def_file = args.file

        # If file path not provided, try the action reference. This only works if the command
        # is executed locally where the content is stored.
        if not wf_def_file:
            action_ref = args.action
            action_manager = self.app.client.managers["Action"]
            action = action_manager.get_by_ref_or_id(ref_or_id=action_ref)

            if not action:
                raise Exception('Unable to identify action "%s".' % action_ref)

            wf_def_file = (
                "/opt/stackstorm/packs/"
                + action.pack
                + "/actions/"
                + action.entry_point
            )

        wf_def = self.get_file_content(wf_def_file)

        return self.manager.inspect(wf_def, **kwargs)

    @add_auth_token_to_kwargs_from_cli
    def run_and_print(self, args, **kwargs):
        errors = self.run(args, **kwargs)

        if not isinstance(errors, list):
            raise TypeError("The inspection result is not type of list: %s" % errors)

        if not errors:
            print("No errors found in workflow definition.")
            return

        print(yaml.safe_dump(errors, default_flow_style=False, allow_unicode=True))



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import dateutil.tz
import dateutil.parser
from pytz import timezone as TZ

from st2client.config import get_config

__all__ = ["parse", "format_isodate"]


def add_utc_tz(dt):
    return dt.replace(tzinfo=dateutil.tz.tzutc())


def parse(value):
    dt = dateutil.parser.parse(str(value))
    return dt if dt.tzinfo else add_utc_tz(dt)


def format_dt(dt):
    """
    Format datetime object for human friendly representation.
    """
    value = dt.strftime("%a, %d %b %Y %H:%M:%S %Z")
    return value


def format_isodate(value, timezone=None):
    """
    Make a ISO date time string human friendly.

    :type value: ``str``

    :rtype: ``str``
    """
    if not value:
        return ""

    # For some reason pylint thinks it returns a tuple but it returns a datetime object
    dt = dateutil.parser.parse(str(value))

    if timezone:
        dt = dt.astimezone(TZ(timezone))

    value = format_dt(dt)
    return value


def format_isodate_for_user_timezone(value):
    """
    Format the provided ISO date time string for human friendly display taking into user timezone
    specific in the config.
    """
    config = get_config()
    timezone = config.get("cli", {}).get("timezone", "UTC")
    result = format_isodate(value=value, timezone=timezone)
    return result



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os
import getpass

import unittest2
import mock

from st2client.base import BaseCLIApp

USER = getpass.getuser()


class BaseCLIAppTestCase(unittest2.TestCase):
    @mock.patch("os.path.isfile", mock.Mock())
    def test_cli_config_file_path(self):
        app = BaseCLIApp()
        args = mock.Mock()

        # 1. Absolute path
        args.config_file = "/tmp/full/abs/path/config.ini"
        result = app._get_config_file_path(args=args)
        self.assertEqual(result, args.config_file)

        args.config_file = "/home/user/st2/config.ini"
        result = app._get_config_file_path(args=args)
        self.assertEqual(result, args.config_file)

        # 2. Path relative to user home directory, should get expanded
        args.config_file = "~/.st2/config.ini"
        result = app._get_config_file_path(args=args)
        expected = os.path.join(os.path.expanduser("~" + USER), ".st2/config.ini")
        self.assertEqual(result, expected)

        # 3. Relative path (should get converted to absolute one)
        args.config_file = "config.ini"
        result = app._get_config_file_path(args=args)
        expected = os.path.join(os.getcwd(), "config.ini")
        self.assertEqual(result, expected)

        args.config_file = ".st2/config.ini"
        result = app._get_config_file_path(args=args)
        expected = os.path.join(os.getcwd(), ".st2/config.ini")
        self.assertEqual(result, expected)



# Copyright 2020 The StackStorm Authors.
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import json
import logging

from os.path import join as pjoin

import six

from st2client.commands import resource
from st2client.commands.noop import NoopCommand
from st2client.formatters import table
from st2client.models.keyvalue import KeyValuePair
from st2client.utils.date import format_isodate_for_user_timezone

LOG = logging.getLogger(__name__)

DEFAULT_LIST_SCOPE = "all"
DEFAULT_GET_SCOPE = "system"
DEFAULT_CUD_SCOPE = "system"


class KeyValuePairBranch(resource.ResourceBranch):
    def __init__(self, description, app, subparsers, parent_parser=None):
        super(KeyValuePairBranch, self).__init__(
            KeyValuePair,
            description,
            app,
            subparsers,
            parent_parser=parent_parser,
            commands={
                "list": KeyValuePairListCommand,
                "get": KeyValuePairGetCommand,
                "delete": KeyValuePairDeleteCommand,
                "create": NoopCommand,
                "update": NoopCommand,
            },
        )

        # Registers extended commands
        self.commands["set"] = KeyValuePairSetCommand(
            self.resource, self.app, self.subparsers
        )
        self.commands["load"] = KeyValuePairLoadCommand(
            self.resource, self.app, self.subparsers
        )
        self.commands["delete_by_prefix"] = KeyValuePairDeleteByPrefixCommand(
            self.resource, self.app, self.subparsers
        )

        # Remove unsupported commands
        # TODO: Refactor parent class and make it nicer
        del self.commands["create"]
        del self.commands["update"]


class KeyValuePairListCommand(resource.ResourceTableCommand):
    display_attributes = [
        "name",
        "value",
        "secret",
        "encrypted",
        "scope",
        "user",
        "expire_timestamp",
    ]
    attribute_transform_functions = {
        "expire_timestamp": format_isodate_for_user_timezone,
    }

    def __init__(self, resource, *args, **kwargs):

        self.default_limit = 50

        super(KeyValuePairListCommand, self).__init__(
            resource,
            "list",
            "Get the list of the %s most recent %s."
            % (self.default_limit, resource.get_plural_display_name().lower()),
            *args,
            **kwargs,
        )
        self.resource_name = resource.get_plural_display_name().lower()
        # Filter options
        self.parser.add_argument(
            "--prefix",
            help=(
                "Only return values with names starting with " "the provided prefix."
            ),
        )
        self.parser.add_argument(
            "-d",
            "--decrypt",
            action="store_true",
            help="Decrypt secrets and displays plain text.",
        )
        self.parser.add_argument(
            "-s",
            "--scope",
            default=DEFAULT_LIST_SCOPE,
            dest="scope",
            help='Scope item is under. Example: "user".',
        )
        self.parser.add_argument(
            "-u",
            "--user",
            dest="user",
            default=None,
            help="User for user scoped items (admin only).",
        )
        self.parser.add_argument(
            "-n",
            "--last",
            type=int,
            dest="last",
            default=self.default_limit,
            help=(
                "List N most recent %s. Use -n -1 to fetch the full result \
                                       set."
                % self.resource_name
            ),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        # Filtering options
        if args.prefix:
            kwargs["prefix"] = args.prefix

        decrypt = getattr(args, "decrypt", False)
        kwargs["params"] = {"decrypt": str(decrypt).lower()}
        scope = getattr(args, "scope", DEFAULT_LIST_SCOPE)
        kwargs["params"]["scope"] = scope
        if args.user:
            kwargs["params"]["user"] = args.user
        kwargs["params"]["limit"] = args.last

        return self.manager.query_with_count(**kwargs)

    @resource.add_auth_token_to_kwargs_from_cli
    def run_and_print(self, args, **kwargs):
        instances, count = self.run(args, **kwargs)
        if args.json or args.yaml:
            self.print_output(
                reversed(instances),
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
                json=args.json,
                yaml=args.yaml,
                attribute_transform_functions=self.attribute_transform_functions,
            )
        else:
            self.print_output(
                instances,
                table.MultiColumnTable,
                attributes=args.attr,
                widths=args.width,
                attribute_transform_functions=self.attribute_transform_functions,
            )

            if args.last and count and count > args.last:
                table.SingleRowTable.note_box(self.resource_name, args.last)


class KeyValuePairGetCommand(resource.ResourceGetCommand):
    pk_argument_name = "name"
    display_attributes = [
        "name",
        "value",
        "secret",
        "encrypted",
        "scope",
        "expire_timestamp",
    ]

    def __init__(self, kv_resource, *args, **kwargs):
        super(KeyValuePairGetCommand, self).__init__(kv_resource, *args, **kwargs)
        self.parser.add_argument(
            "-d",
            "--decrypt",
            action="store_true",
            help="Decrypt secret if encrypted and show plain text.",
        )
        self.parser.add_argument(
            "-s",
            "--scope",
            default=DEFAULT_GET_SCOPE,
            dest="scope",
            help='Scope item is under. Example: "user".',
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        decrypt = getattr(args, "decrypt", False)
        scope = getattr(args, "scope", DEFAULT_GET_SCOPE)
        kwargs["params"] = {"decrypt": str(decrypt).lower()}
        kwargs["params"]["scope"] = scope

        resource_ids = getattr(args, self.pk_argument_name, None)
        resources = self._get_multiple_resources(
            resource_ids=resource_ids, kwargs=kwargs
        )
        return resources


class KeyValuePairSetCommand(resource.ResourceCommand):
    display_attributes = ["name", "value", "scope", "expire_timestamp"]

    def __init__(self, resource, *args, **kwargs):
        super(KeyValuePairSetCommand, self).__init__(
            resource,
            "set",
            "Set an existing %s." % resource.get_display_name().lower(),
            *args,
            **kwargs,
        )

        # --encrypt and --encrypted options are mutually exclusive.
        # --encrypt implies provided value is plain text and should be encrypted whereas
        # --encrypted implies value is already encrypted and should be treated as-is.
        encryption_group = self.parser.add_mutually_exclusive_group()
        encryption_group.add_argument(
            "-e",
            "--encrypt",
            dest="secret",
            action="store_true",
            help="Encrypt value before saving.",
        )
        encryption_group.add_argument(
            "--encrypted",
            dest="encrypted",
            action="store_true",
            help=(
                "Value provided is already encrypted with the "
                "instance crypto key and should be stored as-is."
            ),
        )

        self.parser.add_argument(
            "name", metavar="name", help="Name of the key value pair."
        )
        self.parser.add_argument("value", help="Value paired with the key.", nargs="?")
        self.parser.add_argument(
            "-l",
            "--ttl",
            dest="ttl",
            type=int,
            default=None,
            help="TTL (in seconds) for this value.",
        )
        self.parser.add_argument(
            "-s",
            "--scope",
            dest="scope",
            default=DEFAULT_CUD_SCOPE,
            help="Specify the scope under which you want " + "to place the item.",
        )
        self.parser.add_argument(
            "-u",
            "--user",
            dest="user",
            default=None,
            help="User for user scoped items (admin only).",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        instance = KeyValuePair()
        instance.id = args.name  # TODO: refactor and get rid of id
        instance.name = args.name
        instance.scope = args.scope
        instance.user = args.user

        if not args.value:
            instance.value = input("Please insert value for key: ")
        else:
            instance.value = args.value

        if args.secret:
            instance.secret = args.secret

        if args.encrypted:
            instance.encrypted = args.encrypted

        if args.ttl:
            instance.ttl = args.ttl

        return self.manager.update(instance, **kwargs)

    def run_and_print(self, args, **kwargs):
        instance = self.run(args, **kwargs)
        self.print_output(
            instance,
            table.PropertyValueTable,
            attributes=self.display_attributes,
            json=args.json,
            yaml=args.yaml,
        )


class KeyValuePairDeleteCommand(resource.ResourceDeleteCommand):
    pk_argument_name = "name"

    def __init__(self, resource, *args, **kwargs):
        super(KeyValuePairDeleteCommand, self).__init__(resource, *args, **kwargs)

        self.parser.add_argument(
            "-s",
            "--scope",
            dest="scope",
            default=DEFAULT_CUD_SCOPE,
            help="Specify the scope under which you want " + "to place the item.",
        )
        self.parser.add_argument(
            "-u",
            "--user",
            dest="user",
            default=None,
            help="User for user scoped items (admin only).",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        resource_id = getattr(args, self.pk_argument_name, None)
        scope = getattr(args, "scope", DEFAULT_CUD_SCOPE)
        kwargs["params"] = {}
        kwargs["params"]["scope"] = scope
        kwargs["params"]["user"] = args.user
        instance = self.get_resource(resource_id, **kwargs)

        if not instance:
            raise resource.ResourceNotFoundError(
                'KeyValuePair with id "%s" not found' % resource_id
            )

        instance.id = resource_id  # TODO: refactor and get rid of id
        self.manager.delete(instance, **kwargs)


class KeyValuePairDeleteByPrefixCommand(resource.ResourceCommand):
    """
    Commands which delete all the key value pairs which match the provided
    prefix.
    """

    def __init__(self, resource, *args, **kwargs):
        super(KeyValuePairDeleteByPrefixCommand, self).__init__(
            resource,
            "delete_by_prefix",
            "Delete KeyValue pairs which \
                                                                 match the provided prefix",
            *args,
            **kwargs,
        )

        self.parser.add_argument(
            "-p",
            "--prefix",
            required=True,
            help="Name prefix (e.g. twitter.TwitterSensor:)",
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        prefix = args.prefix
        key_pairs = self.manager.get_all(prefix=prefix)

        to_delete = []
        for key_pair in key_pairs:
            key_pair.id = key_pair.name
            to_delete.append(key_pair)

        deleted = []
        for key_pair in to_delete:
            self.manager.delete(instance=key_pair, **kwargs)
            deleted.append(key_pair)

        return deleted

    def run_and_print(self, args, **kwargs):
        # TODO: Need to use args, instead of kwargs (args=) because of bad API
        # FIX ME
        deleted = self.run(args, **kwargs)
        key_ids = [key_pair.id for key_pair in deleted]

        print("Deleted %s keys" % (len(deleted)))
        print("Deleted key ids: %s" % (", ".join(key_ids)))


class KeyValuePairLoadCommand(resource.ResourceCommand):
    pk_argument_name = "name"
    display_attributes = ["name", "value"]

    def __init__(self, resource, *args, **kwargs):
        help_text = (
            "Load a list of %s from file." % resource.get_plural_display_name().lower()
        )
        super(KeyValuePairLoadCommand, self).__init__(
            resource, "load", help_text, *args, **kwargs
        )

        self.parser.add_argument(
            "-c",
            "--convert",
            action="store_true",
            help=(
                "Convert non-string types (hash, array, boolean,"
                " int, float) to a JSON string before loading it"
                " into the datastore."
            ),
        )
        self.parser.add_argument(
            "file",
            help=(
                "JSON/YAML file containing the %s(s) to load"
                % resource.get_plural_display_name().lower()
            ),
        )

    @resource.add_auth_token_to_kwargs_from_cli
    def run(self, args, **kwargs):
        # normalize the file path to allow for relative files to be specified
        file_path = os.path.normpath(pjoin(os.getcwd(), args.file))

        # load the data (JSON/YAML) from the file
        kvps = resource.load_meta_file(file_path)

        instances = []
        # bail out if file was empty
        if not kvps:
            return instances

        # if the data is not a list (ie. it's a single entry)
        # then make it a list so our process loop is generic
        if not isinstance(kvps, list):
            kvps = [kvps]

        for item in kvps:
            # parse required KeyValuePair properties
            name = item["name"]
            value = item["value"]

            # parse optional KeyValuePair properties
            scope = item.get("scope", DEFAULT_CUD_SCOPE)
            user = item.get("user", None)
            encrypted = item.get("encrypted", False)
            secret = item.get("secret", False)
            ttl = item.get("ttl", None)

            # if the value is not a string, convert it to JSON
            # all keys in the datastore must strings
            if not isinstance(value, six.string_types):
                if args.convert:
                    value = json.dumps(value)
                else:
                    raise ValueError(
                        (
                            "Item '%s' has a value that is not a string."
                            " Either pass in the -c/--convert option to convert"
                            " non-string types to JSON strings automatically, or"
                            " convert the data to a string in the file"
                        )
                        % name
                    )

            # create the KeyValuePair instance
            instance = KeyValuePair()
            instance.id = name  # TODO: refactor and get rid of id
            instance.name = name
            instance.value = value
            instance.scope = scope

            if user:
                instance.user = user
            if encrypted:
                instance.encrypted = encrypted
            if secret:
                instance.secret = secret
            if ttl:
                instance.ttl = ttl

            # encrypted=True and secret=True implies that the value is already encrypted and should
            # be used as such
            if encrypted and secret:
                instance.encrypted = True

            # call the API to create/update the KeyValuePair
            self.manager.update(instance, **kwargs)
            instances.append(instance)

        return instances

    def run_and_print(self, args, **kwargs):
        instances = self.run(args, **kwargs)
        self.print_output(
            instances,
            table.MultiColumnTable,
            attributes=["name", "value", "secret", "scope", "user", "ttl"],
            json=args.json,
            yaml=args.yaml,
        )
